
from asyauth.protocols.ntlm.structures.avpair import AVPAIRType
import datetime
import json

NTLMSERVERINFO_TSV_HDR = ['domainname', 'computername', 'dnsforestname', 'dnscomputername', 'dnsdomainname', 'local_time', 'os_major_version', 'os_minor_version', 'os_build', 'os_guess' ]


import datetime
import io

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/2c57429b-fdd4-488f-b5fc-9e4cf020fcdf
class FILETIME:
	def __init__(self):
		self.dwLowDateTime = None
		self.dwHighDateTime = None
		
		self.datetime = None
	@staticmethod
	def from_bytes(data):
		return FILETIME.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_datetime(dt):
		t = FILETIME()
		# Convert to Windows FILETIME: 100-nanosecond intervals since January 1, 1601 UTC
		# Unix epoch starts at January 1, 1970, so we add the difference
		unix_timestamp = int(dt.timestamp())
		# Convert to 100-nanosecond intervals and add offset from 1601 to 1970
		filetime_value = (unix_timestamp * 10000000) + 116444736000000000
		t.dwLowDateTime = filetime_value & 0xFFFFFFFF
		t.dwHighDateTime = (filetime_value >> 32) & 0xFFFFFFFF
		t.calc_dt()
		return t

	def to_bytes(self):
		if self.dwLowDateTime is None or self.dwHighDateTime is None:
			raise ValueError("FILETIME values cannot be None for serialization")
		return self.dwLowDateTime.to_bytes(4, 'little') + self.dwHighDateTime.to_bytes(4, 'little')
	
	def calc_dt(self):
		if self.dwHighDateTime == 4294967295 and self.dwLowDateTime == 4294967295:
			self.datetime = datetime.datetime(3000, 1, 1, 0, 0)
		else:
			ft = (self.dwHighDateTime << 32) + self.dwLowDateTime
			if ft == 0:
				self.datetime = datetime.datetime(1970, 1, 1, 0, 0)
			else:
				self.datetime = datetime.datetime.utcfromtimestamp((ft - 116444736000000000) / 10000000)
	
	@staticmethod
	def from_dict(d):
		t = FILETIME()
		t.dwLowDateTime = d['dwLowDateTime']
		t.dwHighDateTime = d['dwHighDateTime']
		t.calc_dt()
		return t

	@staticmethod
	def from_buffer(buff):
		t = FILETIME()
		t.dwLowDateTime = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		t.dwHighDateTime = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		t.calc_dt()
		return t



class NTLMServerInfo:
	def __init__(self):
		self.domainname = None
		self.computername = None
		self.dnscomputername = None
		self.dnsdomainname = None
		self.local_time = None
		self.dnsforestname = None
		self.os_major_version = None
		self.os_minor_version = None
		self.os_build = None
		self.os_guess = None
		self.creds = None
	
	@staticmethod
	def from_challenge(challenge):
		si = NTLMServerInfo()
		ti = challenge.TargetInfo
		for k in ti:
			if k == AVPAIRType.MsvAvNbDomainName:
				si.domainname = ti[k]
			elif k == AVPAIRType.MsvAvNbComputerName:
				si.computername = ti[k]
			elif k == AVPAIRType.MsvAvDnsDomainName:
				si.dnsdomainname = ti[k]
			elif k == AVPAIRType.MsvAvDnsComputerName:
				si.dnscomputername = ti[k]
			elif k == AVPAIRType.MsvAvDnsTreeName:
				si.dnsforestname = ti[k]
			elif k == AVPAIRType.MsvAvTimestamp:
				if isinstance(ti[k], bytes):
					si.local_time = FILETIME.from_bytes(ti[k]).datetime
				elif isinstance(ti[k], datetime):
					si.local_time = ti[k]
		
		if challenge.Version is not None:
			if challenge.Version.ProductMajorVersion is not None:
				si.os_major_version = challenge.Version.ProductMajorVersion
			if challenge.Version.ProductMinorVersion is not None:
				si.os_minor_version = challenge.Version.ProductMinorVersion
			if challenge.Version.ProductBuild is not None:
				si.os_build = challenge.Version.ProductBuild
			if challenge.Version.WindowsProduct is not None:
				si.os_guess = challenge.Version.WindowsProduct
				
		return si

	def to_dict(self):
		t = {
			'domainname' : self.domainname,
			'computername' : self.computername,
			'dnscomputername' : self.dnscomputername,
			'dnsdomainname' : self.dnsdomainname,
			'local_time' : self.local_time,
			'dnsforestname' : self.dnsforestname,
			'os_build' : self.os_build,
			'os_guess' : self.os_guess,
			'os_major_version' : None,
			'os_minor_version' : None,
		}
		if self.os_major_version is not None:
			t['os_major_version'] = self.os_major_version.name
		if self.os_minor_version is not None:
			t['os_minor_version'] = self.os_minor_version.name
		if self.creds is not None:
			t['creds'] = self.creds.to_dict()
		return t

	def to_tsv(self, separator = '\t'):
		def vn(x):
			if x is None:
				return ''
			return str(x)

		d = self.to_dict()
		return separator.join([ vn(d[x]) for x in NTLMSERVERINFO_TSV_HDR])
		
	def __str__(self):
		t = '=== Server Info ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k]) 
			
		return t

	def to_json(self):
		return json.dumps(self.to_dict())

	def to_grep(self):
		t  = ''
		t += '[domainname,%s]' % self.domainname
		t += '[computername,%s]' %  self.computername
		t += '[dnscomputername,%s]' %  self.dnscomputername
		t += '[dnsdomainname,%s]' %  self.dnsdomainname
		t += '[dnsforestname,%s]' %  self.dnsforestname
		t += '[os_build,%s]' %  self.os_build
		t += '[os_guess,%s]' %  self.os_guess
		if self.local_time is not None:
			t += '[local_time,%s]' %  self.local_time.isoformat()
		if self.os_major_version is not None:
			t += '[os_major,%s]' % self.os_major_version.value
		if self.os_minor_version is not None:
			t += '[os_minor,%s]' % self.os_minor_version.value
		
		return t


import enum



class SSPIResult(enum.Enum):
	OK = 'OK'
	CONTINUE = 'CONT'
	ERR = 'ERR'

class ISC_REQ(enum.IntFlag):
	DELEGATE = 1
	MUTUAL_AUTH = 2
	REPLAY_DETECT = 4
	SEQUENCE_DETECT = 8
	CONFIDENTIALITY = 16
	USE_SESSION_KEY = 32
	PROMPT_FOR_CREDS = 64
	USE_SUPPLIED_CREDS = 128
	ALLOCATE_MEMORY = 256
	USE_DCE_STYLE = 512
	DATAGRAM = 1024
	CONNECTION = 2048
	CALL_LEVEL = 4096
	FRAGMENT_SUPPLIED = 8192
	EXTENDED_ERROR = 16384
	STREAM = 32768
	INTEGRITY = 65536
	IDENTIFY = 131072
	NULL_SESSION = 262144
	MANUAL_CRED_VALIDATION = 524288
	RESERVED1 = 1048576
	FRAGMENT_TO_FIT = 2097152
	HTTP = 0x10000000

class SEC_E(enum.Enum):
	OK = 0x00000000 
	CONTINUE_NEEDED = 0x00090312 
	INSUFFICIENT_MEMORY = 0x80090300 #Not enough memory is available to complete this request.
	INVALID_HANDLE = 0x80090301 #The handle specified is invalid.
	UNSUPPORTED_FUNCTION = 0x80090302 #The function requested is not supported.
	TARGET_UNKNOWN = 0x80090303 #The specified target is unknown or unreachable.
	INTERNAL_ERROR = 0x80090304 #The Local Security Authority (LSA) cannot be contacted.
	SECPKG_NOT_FOUND = 0x80090305  #The requested security package does not exist.
	NOT_OWNER = 0x80090306  #The caller is not the owner of the desired credentials.
	CANNOT_INSTALL = 0x80090307  #The security package failed to initialize and cannot be installed.
	INVALID_TOKEN = 0x80090308  #The token supplied to the function is invalid.
	CANNOT_PACK = 0x80090309  #The security package is not able to marshal the logon buffer, so the logon attempt has failed.
	QOP_NOT_SUPPORTED = 0x8009030A  #The per-message quality of protection is not supported by the security package.
	NO_IMPERSONATION = 0x8009030B  #The security context does not allow impersonation of the client.
	LOGON_DENIED = 0x8009030C  #The logon attempt failed.
	UNKNOWN_CREDENTIALS = 0x8009030D  #The credentials supplied to the package were not recognized.
	NO_CREDENTIALS = 0x8009030E  #No credentials are available in the security package.
	MESSAGE_ALTERED = 0x8009030F  #The message or signature supplied for verification has been altered.
	OUT_OF_SEQUENCE = 0x80090310  #The message supplied for verification is out of sequence.
	NO_AUTHENTICATING_AUTHORITY = 0x80090311  #No authority could be contacted for authentication.
	BAD_PKGID = 0x80090316  #The requested security package does not exist.
	CONTEXT_EXPIRED = 0x80090317  #The context has expired and can no longer be used.
	INCOMPLETE_MESSAGE = 0x80090318  #The supplied message is incomplete. The signature was not verified.
	INCOMPLETE_CREDENTIALS = 0x80090320  #The credentials supplied were not complete and could not be verified. The context could not be initialized.
	BUFFER_TOO_SMALL = 0x80090321  #The buffers supplied to a function was too small.
	WRONG_PRINCIPAL = 0x80090322  #The target principal name is incorrect.
	TIME_SKEW = 0x80090324  #The clocks on the client and server machines are skewed.
	UNTRUSTED_ROOT = 0x80090325  #The certificate chain was issued by an authority that is not trusted.
	ILLEGAL_MESSAGE = 0x80090326  #The message received was unexpected or badly formatted.
	CERT_UNKNOWN = 0x80090327  #An unknown error occurred while processing the certificate.
	CERT_EXPIRED = 0x80090328  # The received certificate has expired.
	ENCRYPT_FAILURE = 0x80090329  #The specified data could not be encrypted.
	DECRYPT_FAILURE = 0x80090330  #The specified data could not be decrypted.
	ALGORITHM_MISMATCH = 0x80090331  #The client and server cannot communicate because they do not possess a common algorithm.
	SECURITY_QOS_FAILED = 0x80090332  #The security context could not be established due to a failure in the requested quality of service (for example, mutual authentication or delegation).
	UNFINISHED_CONTEXT_DELETED = 0x80090333  #A security context was deleted before the context was completed. This is considered a logon failure.
	NO_TGT_REPLY = 0x80090334  #The client is trying to negotiate a context and the server requires user-to-user but did not send a ticket granting ticket (TGT) reply.
	NO_IP_ADDRESSES = 0x80090335  #Unable to accomplish the requested task because the local machine does not have an IP addresses.
	WRONG_CREDENTIAL_HANDLE = 0x80090336  #The supplied credential handle does not match the credential associated with the security context.
	CRYPTO_SYSTEM_INVALID = 0x80090337  #The cryptographic system or checksum function is invalid because a required function is unavailable.
	MAX_REFERRALS_EXCEEDED = 0x80090338  #The number of maximum ticket referrals has been exceeded.
	MUST_BE_KDC = 0x80090339  #The local machine must be a Kerberos domain controller (KDC), and it is not.
	STRONG_CRYPTO_NOT_SUPPORTED = 0x8009033A  #The other end of the security negotiation requires strong cryptographics, but it is not supported on the local machine.
	TOO_MANY_PRINCIPALS = 0x8009033B  #The KDC reply contained more than one principal name.
	NO_PA_DATA = 0x8009033C  #Expected to find PA data for a hint of what etype to use, but it was not found.
	PKINIT_NAME_MISMATCH = 0x8009033D  #The client certificate does not contain a valid user principal name (UPN), or does not match the client name in the logon request. Contact your administrator.
	SMARTCARD_LOGON_REQUIRED = 0x8009033E  #Smart card logon is required and was not used.
	SHUTDOWN_IN_PROGRESS = 0x8009033F  #A system shutdown is in progress.
	KDC_INVALID_REQUEST = 0x80090340  #An invalid request was sent to the KDC.
	KDC_UNABLE_TO_REFER = 0x80090341  #The KDC was unable to generate a referral for the service requested.
	KDC_UNKNOWN_ETYPE = 0x80090342  #The encryption type requested is not supported by the KDC.
	UNSUPPORTED_PREAUTH = 0x80090343  #An unsupported pre-authentication mechanism was presented to the Kerberos package.
	DELEGATION_REQUIRED = 0x80090345  #The requested operation cannot be completed. The computer must be trusted for delegation, and the current user account must be configured to allow delegation.
	BAD_BINDINGS = 0x80090346  #Client's supplied Security Support Provider Interface (SSPI) channel bindings were incorrect.
	MULTIPLE_ACCOUNTS = 0x80090347  #The received certificate was mapped to multiple accounts.
	NO_KERB_KEY = 0x80090348  #No Kerberos key was found.
	CERT_WRONG_USAGE = 0x80090349  #The certificate is not valid for the requested usage.
	DOWNGRADE_DETECTED = 0x80090350  #The system detected a possible attempt to compromise security. Ensure that you can contact the server that authenticated you.
	SMARTCARD_CERT_REVOKED = 0x80090351  #The smart card certificate used for authentication has been revoked. Contact your system administrator. The event log might contain additional information.
	ISSUING_CA_UNTRUSTED = 0x80090352  #An untrusted certification authority (CA) was detected while processing the smart card certificate used for authentication. Contact your system administrator.
	REVOCATION_OFFLINE_C = 0x80090353  #The revocation status of the smart card certificate used for authentication could not be determined. Contact your system administrator.
	PKINIT_CLIENT_FAILURE = 0x80090354  #The smart card certificate used for authentication was not trusted. Contact your system administrator.
	SMARTCARD_CERT_EXPIRED = 0x80090355  #The smart card certificate used for authentication has expired. Contact your system administrator.
	NO_S4U_PROT_SUPPORT = 0x80090356  #The Kerberos subsystem encountered an error. A service for user protocol requests was made against a domain controller that does not support services for users.
	CROSSREALM_DELEGATION_FAILURE = 0x80090357  #An attempt was made by this server to make a Kerberos-constrained delegation request for a target outside the server's realm. This is not supported and indicates a misconfiguration on this server's allowed-to-delegate-to list. Contact your administrator.
	REVOCATION_OFFLINE_KDC = 0x80090358  #The revocation status of the domain controller certificate used for smart card authentication could not be determined. The system event log contains additional information. Contact your system administrator.
	ISSUING_CA_UNTRUSTED_KDC = 0x80090359  #An untrusted CA was detected while processing the domain controller certificate used for authentication. The system event log contains additional information. Contact your system administrator.
	KDC_CERT_EXPIRED = 0x8009035A  #The domain controller certificate used for smart card logon has expired. Contact your system administrator with the contents of your system event log.
	KDC_CERT_REVOKED = 0x8009035B  #The domain controller certificate used for smart card logon has been revoked. Contact your system administrator with the contents of your system event log.
	INVALID_PARAMETER = 0x8009035D  #One or more of the parameters passed to the function were invalid.
	DELEGATION_POLICY = 0x8009035E  #The client policy does not allow credential delegation to the target server.
	POLICY_NLTM_ONLY = 0x8009035F  #The client policy does not allow credential delegation to the target server with NLTM only authentication.
	RENEGOTIATE = 590625
	COMPLETE_AND_CONTINUE = 590612
	COMPLETE_NEEDED = 590611
	#INCOMPLETE_CREDENTIALS = 590624

class SECPKG_CRED(enum.IntFlag):
	AUTOLOGON_RESTRICTED = 0x00000010 	#The security does not use default logon credentials or credentials from Credential Manager.
										#This value is supported only by the Negotiate security package.
										#Windows Server 2008, Windows Vista, Windows Server 2003 and Windows XP:  This value is not supported.

	BOTH = 3							#Validate an incoming credential or use a local credential to prepare an outgoing token. This flag enables both other flags. This flag is not valid with the Digest and Schannel SSPs.
	INBOUND = 1							#Validate an incoming server credential. Inbound credentials might be validated by using an authenticating authority when InitializeSecurityContext (General) or AcceptSecurityContext (General) is called. If such an authority is not available, the function will fail and return SEC_E_NO_AUTHENTICATING_AUTHORITY. Validation is package specific.
	OUTBOUND = 2						#Allow a local client credential to prepare an outgoing token.
	PROCESS_POLICY_ONLY = 0x00000020 	#The function processes server policy and returns SEC_E_NO_CREDENTIALS, indicating that the application should prompt for credentials.
										#This value is supported only by the Negotiate security package.
										#Windows Server 2008, Windows Vista, Windows Server 2003 and Windows XP:  This value is not supported.

# https://docs.microsoft.com/en-us/windows/desktop/api/sspi/ns-sspi-_secbuffer
class SECBUFFER_TYPE(enum.Enum):
	SECBUFFER_ATTRMASK = 4026531840 #The buffer contains a bitmask for a SECBUFFER_READONLY_WITH_CHECKSUM buffer.
	
	SECBUFFER_EMPTY = 0 #This is a placeholder in the buffer array. The caller can supply several such entries in the array, and the security package can return information in them. For more information, see SSPI Context Semantics.
	SECBUFFER_DATA = 1 #The buffer contains common data. The security package can read and write this data, for example, to encrypt some or all of it.
	SECBUFFER_TOKEN = 2 #The buffer contains the security token portion of the message. This is read-only for input parameters or read/write for output parameters.
	SECBUFFER_PKG_PARAMS = 3 #These are transport-to-package–specific parameters. For example, the NetWare redirector may supply the server object identifier, while DCE RPC can supply an association UUID, and so on.
	SECBUFFER_MISSING = 4 #The security package uses this value to indicate the number of missing bytes in a particular message. The pvBuffer member is ignored in this type.
	SECBUFFER_EXTRA = 5 #The security package uses this value to indicate the number of extra or unprocessed bytes in a message.
	SECBUFFER_STREAM_TRAILER = 6 #The buffer contains a protocol-specific trailer for a particular record. It is not usually of interest to callers.
	SECBUFFER_STREAM_HEADER = 7 #The buffer contains a protocol-specific header for a particular record. It is not usually of interest to callers.
	SECBUFFER_NEGOTIATION_INFO = 8
	SECBUFFER_PADDING = 9 
	SECBUFFER_STREAM = 10
	SECBUFFER_MECHLIST = 11 #The buffer contains a protocol-specific list of object identifiers (OIDs). It is not usually of interest to callers.
	SECBUFFER_MECHLIST_SIGNATURE = 12 #The buffer contains a signature of a SECBUFFER_MECHLIST buffer. It is not usually of interest to callers.
	SECBUFFER_TARGET = 13 #This flag is reserved. Do not use it.
	SECBUFFER_CHANNEL_BINDINGS = 14  #	The buffer contains channel binding information.
	SECBUFFER_CHANGE_PASS_RESPONSE = 15 #The buffer contains a DOMAIN_PASSWORD_INFORMATION structure.
	SECBUFFER_TARGET_HOST = 16 #The buffer specifies the service principal name (SPN) of the target.
								#This value is supported by the Digest security package when used with channel bindings.
								#Windows Server 2008, Windows Vista, Windows Server 2003 and Windows XP:  This value is not supported.
	SECBUFFER_ALERT = 17 #The buffer contains an alert message.
	SECBUFFER_APPLICATION_PROTOCOLS = 18 #The buffer contains a list of application protocol IDs, one list per application protocol negotiation extension type to be enabled.
	SECBUFFER_SRTP_PROTECTION_PROFILES = 19 #The buffer contains the list of SRTP protection profiles, in descending order of preference.
	SECBUFFER_SRTP_MASTER_KEY_IDENTIFIER = 20 #The buffer contains the SRTP master key identifier.
	SECBUFFER_TOKEN_BINDING = 21 #The buffer contains the supported token binding protocol version and key parameters, in descending order of preference.
	SECBUFFER_PRESHARED_KEY = 22 #The buffer contains the preshared key. The maximum allowed PSK buffer size is 256 bytes.
	SECBUFFER_PRESHARED_KEY_IDENTITY = 23 #The buffer contains the preshared key identity.
	SECBUFFER_DTLS_MTU = 24#The buffer contains the setting for the maximum transmission unit (MTU) size for DTLS only. The default value is 1096 and the valid configurable range is between 200 and 64*1024.
	"""
	In addition, BufferType can combine the following flags with any of the flags in the preceding table by using a bitwise-OR operation.
	Value 	Meaning
	SECBUFFER_READONLY
	2147483648 (0x80000000)
		The buffer is read-only with no checksum. This flag is intended for sending header information to the security package for computing the checksum. The package can read this buffer, but cannot modify it.
	SECBUFFER_READONLY_WITH_CHECKSUM
	268435456 (0x10000000)
		The buffer is read-only with a checksum
	"""

class SECPKG_ATTR(enum.Enum):
	SESSION_KEY = 9
	C_ACCESS_TOKEN = 0x80000012 #The pBuffer parameter contains a pointer to a SecPkgContext_AccessToken structure that specifies the access token for the current security context. This attribute is supported only on the server.
	C_FULL_ACCESS_TOKEN = 0x80000082 #The pBuffer parameter contains a pointer to a SecPkgContext_AccessToken structure that specifies the access token for the current security context. This attribute is supported only on the server.
	CERT_TRUST_STATUS = 0x80000084 #The pBuffer parameter contains a pointer to a CERT_TRUST_STATUS structure that specifies trust information about the certificate.This attribute is supported only on the client.
	CREDS = 0x80000080 # The pBuffer parameter contains a pointer to a SecPkgContext_ClientCreds structure that specifies client credentials. The client credentials can be either user name and password or user name and smart card PIN. This attribute is supported only on the server.
	CREDS_2 = 0x80000086 #The pBuffer parameter contains a pointer to a SecPkgContext_ClientCreds structure that specifies client credentials. If the client credential is user name and password, the buffer is a packed KERB_INTERACTIVE_LOGON structure. If the client credential is user name and smart card PIN, the buffer is a packed KERB_CERTIFICATE_LOGON structure. If the client credential is an online identity credential, the buffer is a marshaled SEC_WINNT_AUTH_IDENTITY_EX2 structure. This attribute is supported only on the CredSSP server. Windows Server 2008 R2, Windows 7, Windows Server 2008, Windows Vista, Windows Server 2003 and Windows XP:  This value is not supported.
	NEGOTIATION_PACKAGE = 0x80000081 #The pBuffer parameter contains a pointer to a SecPkgContext_PackageInfo structure that specifies the name of the authentication package negotiated by the Microsoft Negotiate provider.
	PACKAGE_INFO = 10 #The pBuffer parameter contains a pointer to a SecPkgContext_PackageInfostructure.Returns information on the SSP in use.
	SERVER_AUTH_FLAGS = 0x80000083 #The pBuffer parameter contains a pointer to a SecPkgContext_Flags structure that specifies information about the flags in the current security context. This attribute is supported only on the client.
	SIZES = 0x0 #The pBuffer parameter contains a pointer to a SecPkgContext_Sizes structure. Queries the sizes of the structures used in the per-message functions and authentication exchanges.
	SUBJECT_SECURITY_ATTRIBUTES = 124 #	The pBuffer parameter contains a pointer to a SecPkgContext_SubjectAttributes structure. This value returns information about the security attributes for the connection. This value is supported only on the CredSSP server. Windows Server 2008, Windows Vista, Windows Server 2003 and Windows XP:  This value is not supported.
	ENDPOINT_BINDINGS = 26





from minikerberos.gssapi.gssapi import get_gssapi, GSSWrapToken
from minikerberos.protocol.asn1_structs import AP_REP
from minikerberos.protocol.encryption import Key
from asyauth.common.winapi.constants import ISC_REQ
from minikerberos.protocol.encryption import Key, _enctype_table
from minikerberos.protocol.asn1_structs import AP_REP, AP_REQ
from minikerberos.protocol.encryption import Key
from minikerberos.gssapi.gssapi import GSSWrapToken
from asyauth.protocols.kerberos.gssapismb import get_gssapi as gssapi_smb
from asyauth.common.winapi.token import InitialContextToken

from wsnet.pyodide.clientauth import WSNETAuth



class KerberosClientWSNET:
	def __init__(self, settings):
		self.iterations = 0
		self.settings = settings
		self.ksspi = WSNETAuth()
		self.client = None
		self.target = None
		self.gssapi = None
		self.etype = None
		self.session_key = None
		self.seq_number = None
		
	def get_seq_number(self):
		"""
		Returns the initial sequence number. It is 0 by default, but can be adjusted during authentication, 
		by passing the 'seq_number' parameter in the 'authenticate' function
		"""
		return self.seq_number
	
	def signing_needed(self):
		"""
		Checks if integrity protection was negotiated
		"""
		return ISC_REQ.INTEGRITY in self.flags
	
	def encryption_needed(self):
		"""
		Checks if confidentiality flag was negotiated
		"""
		return ISC_REQ.CONFIDENTIALITY in self.flags
				
	async def sign(self, data, message_no, direction = 'init'):
		"""
		Signs a message. 
		"""
		return self.gssapi.GSS_GetMIC(data, message_no, direction = direction)	
		
	async def encrypt(self, data, message_no, *args, **kwargs):
		"""
		Encrypts a message. 
		"""
		data, eeee  = self.gssapi.GSS_Wrap(data, message_no, *args, **kwargs)
		return data, eeee 
		
	async def decrypt(self, data, message_no, *args, **kwargs):
		"""
		Decrypts message. Also performs integrity checking.
		"""
		return self.gssapi.GSS_Unwrap(data, message_no, *args, **kwargs)
	
	def get_session_key(self):
		return self.session_key.contents
	
	async def authenticate(self, authData, flags:ISC_REQ = None, cb_data = None, spn=None, **kwargs):
		try:
			if flags is None:
				flags = ISC_REQ.CONFIDENTIALITY | \
							ISC_REQ.INTEGRITY | \
							ISC_REQ.REPLAY_DETECT | \
							ISC_REQ.SEQUENCE_DETECT #| \
							#ISC_REQ.MUTUAL_AUTH 
							
			
			if ISC_REQ.MUTUAL_AUTH in flags:
				
				if self.iterations == 0:
					flags = ISC_REQ.CONFIDENTIALITY | \
							ISC_REQ.INTEGRITY | \
							ISC_REQ.MUTUAL_AUTH | \
							ISC_REQ.REPLAY_DETECT | \
							ISC_REQ.SEQUENCE_DETECT|\
							ISC_REQ.USE_DCE_STYLE
					
					status, retflags, apreq, err = await self.ksspi.authenticate('KERBEROS', '', spn, 3, flags.value, authdata = b'')
					if err is not None:
						raise err
					self.flags = ISC_REQ(retflags)
					self.iterations += 1
					return apreq, True, None
				
				elif self.iterations == 1:
					status, retflags, data, err = await self.ksspi.authenticate('KERBEROS', '', spn, 3, flags.value, authdata = authData)
					if err is not None:
						return None, None, err
					self.flags = ISC_REQ(retflags)
					session_key_data, err = await self.ksspi.get_sessionkey()
					if err is not None:
						return None, None, err
						
					aprep = AP_REP.load(data).native
					self.session_key = Key(aprep['enc-part']['etype'], session_key_data)

					if ISC_REQ.USE_DCE_STYLE in self.flags:
						self.gssapi = gssapi_smb(self.session_key)
					else:
						self.gssapi = get_gssapi(self.session_key)

					#self.gssapi = get_gssapi(self.session_key)

					if aprep['enc-part']['etype'] != 23: #no need for seq number in rc4
						raw_seq_data, err = await self.ksspi.get_sequenceno()
						if err is not None:
							return None, None, err
						self.seq_number = GSSWrapToken.from_bytes(raw_seq_data[16:]).SND_SEQ
					
					self.iterations += 1
					await self.ksspi.disconnect()
					return data, False, None
					
				else:
					raise Exception('SSPI Kerberos -RPC - auth encountered too many calls for authenticate.')
			
				
			else:
				status, retflags, tokendata, err = await self.ksspi.authenticate('KERBEROS', '', spn, 3, flags.value, authdata = b'')
				if err is not None:
					return None, None, err
				self.flags = ISC_REQ(retflags)
				token = InitialContextToken.load(tokendata)
				apreq = AP_REQ(token.native['innerContextToken'])
				
				session_key_data, err = await self.ksspi.get_sessionkey()
				if err is not None:
					raise err
				
				cipher = _enctype_table[int(apreq.native['ticket']['enc-part']['etype'])]()
				self.session_key = Key(cipher.enctype, session_key_data)
				await self.ksspi.disconnect()

				if apreq.native['ticket']['enc-part']['etype'] != 23: #no need for seq number in rc4
					raw_seq_data, err = await self.ksspi.get_sequenceno()
					if err is not None:
						return None, None, err
					self.seq_number = GSSWrapToken.from_bytes(raw_seq_data[16:]).SND_SEQ

				if ISC_REQ.USE_DCE_STYLE in self.flags:
					self.gssapi = gssapi_smb(self.session_key)
				else:
					self.gssapi = get_gssapi(self.session_key)

				return apreq.dump(), False, None
		except Exception as e:
			return None, None, e
		


import io

from asyauth.utils.timestamp import datetime2timestamp, timestamp2datetime
from asyauth.protocols.ntlm.structures.avpair import AVPairs, AVPAIRType

# https://msdn.microsoft.com/en-us/library/cc236648.aspx
class LMResponse:
	def __init__(self):
		self.Response = None
		
	def to_bytes(self):
		return self.Response

	@staticmethod
	def from_bytes(bbuff):
		return LMResponse.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		t = LMResponse()
		t.Response = buff.read(24)
		return t

	def __repr__(self):
		t  = '== LMResponse ==\r\n'
		t += 'Response: %s\r\n' % repr(self.Response.hex())
		return t


# https://msdn.microsoft.com/en-us/library/cc236649.aspx
class LMv2Response:
	def __init__(self):
		self.Response = None
		self.ChallengeFromClinet = None
		
		
	def to_bytes(self):
		return self.Response + self.ChallengeFromClinet

	@staticmethod
	def from_bytes(bbuff):
		return LMv2Response.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		t = LMv2Response()
		t.Response = buff.read(16).hex()
		t.ChallengeFromClinet = buff.read(8).hex()
		return t

	def __repr__(self):
		t  = '== LMv2Response ==\r\n'
		t += 'Response: %s\r\n' % repr(self.Response)
		t += 'ChallengeFromClinet: %s\r\n' % repr(self.ChallengeFromClinet)
		return t


# https://msdn.microsoft.com/en-us/library/cc236651.aspx
class NTLMv1Response:
	def __init__(self):
		self.Response = None
		
	def to_bytes(self):
		return self.Response

	@staticmethod
	def from_bytes(bbuff):
		return NTLMv1Response.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		t = NTLMv1Response()
		t.Response = buff.read(24).hex()
		return t

	def __repr__(self):
		t  = '== NTLMv1Response ==\r\n'
		t += 'Response: %s\r\n' % repr(self.Response)
		return t


# https://msdn.microsoft.com/en-us/library/cc236653.aspx
class NTLMv2Response:
	def __init__(self):
		self.Response = None
		self.ChallengeFromClinet = None
		
	def to_bytes(self):
		return self.Response + self.ChallengeFromClinet.to_bytes()

	@staticmethod
	def from_bytes(bbuff):
		return NTLMv2Response.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		t = NTLMv2Response()
		t.Response = buff.read(16)
		pos = buff.tell()
		t.ChallengeFromClinet = NTLMv2ClientChallenge.from_buffer(buff)

		return t

	def __repr__(self):
		t  = '== NTLMv2Response ==\r\n'
		t += 'Response           : %s\r\n' % repr(self.Response.hex())
		t += 'ChallengeFromClinet: %s\r\n' % repr(self.ChallengeFromClinet)
		return t


class NTLMv2ClientChallenge:
	def __init__(self):
		self.RespType   = 1
		self.HiRespType = 1
		self.Reserved1  = 0
		self.TimeStamp  = None #bytes! because of conversion error :(
		self.Reserved2  = 0
		self.ChallengeFromClient = None
		self.Reserved3  = 0
		self.Details    = None #named AVPairs in the documentation
		
		self.timestamp_dt = None
		self.raw_data = b''
	
	@staticmethod
	def construct(timestamp, client_challenge, details):
		"""
		timestamp: datetime.datetime
		client_challenge: 8 bytes
		details: AVPairs object
		"""
		cc = NTLMv2ClientChallenge()
		cc.TimeStamp = datetime2timestamp(timestamp)
		cc.ChallengeFromClient = client_challenge
		cc.Details = details
		cc.timestamp_dt = timestamp
		return cc
		
	def to_bytes(self):
		t  = self.RespType.to_bytes(1 , byteorder = 'little', signed = False)
		t += self.HiRespType.to_bytes(1 , byteorder = 'little', signed = False)
		t += self.Reserved1.to_bytes(6, byteorder = 'little', signed = False)
		t += self.TimeStamp
		t += self.ChallengeFromClient
		t += self.Reserved2.to_bytes(4, byteorder = 'little', signed = False)
		t += self.Details.to_bytes()
		t += self.Reserved3.to_bytes(4, byteorder = 'little', signed = False)
		
		return t

	@staticmethod
	def from_bytes(bbuff):
		return NTLMv2ClientChallenge.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		cc = NTLMv2ClientChallenge()
		cc.RespType   = int.from_bytes(buff.read(1), byteorder = 'little', signed = False)
		cc.HiRespType = int.from_bytes(buff.read(1), byteorder = 'little', signed = False)
		cc.Reserved1  = int.from_bytes(buff.read(6), byteorder = 'little', signed = False)
		cc.TimeStamp  = buff.read(8)
		cc.ChallengeFromClient = buff.read(8)
		cc.Reserved2  = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		cc.Details    = AVPairs.from_buffer(buff) #referred to as ServerName in the documentation
		cc.Reserved3 = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		
		cc.timestamp_dt = timestamp2datetime(cc.TimeStamp)

		return cc

	def __repr__(self):
		t  = '== NTLMv2ClientChallenge ==\r\n'
		t += 'RespType           : %s\r\n' % repr(self.RespType)
		t += 'TimeStamp          : %s\r\n' % repr(self.timestamp_dt)
		t += 'ChallengeFromClient: %s\r\n' % repr(self.ChallengeFromClient.hex())
		t += 'Details            : %s\r\n' % repr(self.Details)
		return t

def test():
	import datetime
	test_data = bytes.fromhex('0101000000000000aec600bfc5fdd4011bfa20699d7628730000000002000800540045005300540001001200570049004e003200300031003900410044000400120074006500730074002e0063006f007200700003002600570049004e003200300031003900410044002e0074006500730074002e0063006f007200700007000800aec600bfc5fdd40106000400020000000800300030000000000000000000000000200000527d27f234de743760966384d36f61ae2aa4fc2a380699f8caa600011b486d890a0010000000000000000000000000000000000009001e0063006900660073002f00310030002e00310030002e00310030002e0032000000000000000000')
	
	cc = NTLMv2ClientChallenge.from_bytes(test_data)
	print(repr(cc))
	
	cc2 = NTLMv2ClientChallenge.from_bytes(cc.to_bytes())
	print(repr(cc2))
	print('=== Original ===')
	print(hexdump(test_data))
	print('=== CC ===')
	print(hexdump(cc.to_bytes()))
	
	### assertions here fail because of the timestamp re-conversion loosing info (float-int conversion)
	#assert cc.to_bytes() == test_data
	#assert cc2.to_bytes() == test_data
	
	details = AVPairs({AVPAIRType.MsvAvNbDomainName: 'TEST', AVPAIRType.MsvAvNbComputerName: 'WIN2019AD', AVPAIRType.MsvAvDnsDomainName: 'test.corp', AVPAIRType.MsvAvDnsComputerName: 'WIN2019AD.test.corp', AVPAIRType.MsvAvTimestamp: b'\xae\xc6\x00\xbf\xc5\xfd\xd4\x01', AVPAIRType.MsvAvFlags: b'\x02\x00\x00\x00', AVPAIRType.MsvAvSingleHost: b"0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00 \x00\x00R}'\xf24\xdet7`\x96c\x84\xd3oa\xae*\xa4\xfc*8\x06\x99\xf8\xca\xa6\x00\x01\x1bHm\x89", AVPAIRType.MsvChannelBindings: b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00', AVPAIRType.MsvAvTargetName: 'termsrv/10.10.10.2'})
	timestamp = datetime.datetime(2019,1,1)
	client_challenge = os.urandom(8)
	
	cc3 = NTLMv2ClientChallenge.construct(timestamp, client_challenge, details)
	print(repr(cc3))
	cc4 = NTLMv2ClientChallenge.from_bytes(cc3.to_bytes())

if __name__ == '__main__':
	from asyauth.utils.hexdump import hexdump
	import os
	test()


from typing import List
from asyauth.common.credentials import UniCredential
from asyauth.common.constants import asyauthSecret, asyauthProtocol, asyauthSubProtocol
from asyauth.common.subprotocols import SubProtocol
from asyauth.common.subprotocols import SubProtocolNative

class SPNEGOCredential(UniCredential):
	def __init__(self, credentials:List[UniCredential] = [], subprotocol:SubProtocol = SubProtocolNative()):
		UniCredential.__init__(self, protocol = asyauthProtocol.SPNEGO, subprotocol=subprotocol)
		self.credentials = credentials
	
	def build_context(self, *args, **kwargs):
		if self.subprotocol.type == asyauthSubProtocol.NATIVE:
			from asyauth.protocols.spnego.client.native import SPNEGOClientNative
			if len(self.credentials) == 1 and isinstance(self.credentials[0], SPNEGOClientNative):
				return self.credentials[0].get_copy()
			
			sspi_ctx = SPNEGOClientNative(self)
			for credential in self.credentials:
				context = credential.build_context(**kwargs)
				if credential.protocol == asyauthProtocol.PLAIN:
					# this happens when the developer creates a plain credential and adds it to the SPNEGO credential
					# it's not possible to tell what the protocol is, so we assume it's NTLM
					context = credential.build_context(asyauthProtocol.NTLM)
					sspi_ctx.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', context)

				elif credential.protocol == asyauthProtocol.KERBEROS:
					sspi_ctx.add_auth_context('MS KRB5 - Microsoft Kerberos 5', context)
				elif credential.protocol == asyauthProtocol.NTLM:
					sspi_ctx.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', context)
				else:
					raise Exception('Authentication protocol "%s" is not supported for SPNEGO' % credential.protocol)
			return sspi_ctx

		else:
			raise Exception('Unsupported subprotocol "%s"' % self.subprotocol)


from typing import List
from asyauth.common.credentials import UniCredential
from asyauth.common.constants import asyauthSecret, asyauthProtocol, asyauthSubProtocol
from asyauth.common.subprotocols import SubProtocol
from asyauth.common.subprotocols import SubProtocolNative

class SPNEGOEXCredential(UniCredential):
	def __init__(self, credentials:List[UniCredential] = [], subprotocol:SubProtocol = SubProtocolNative()):
		UniCredential.__init__(self, protocol = asyauthProtocol.SPNEGOEX, subprotocol=subprotocol)
		self.credentials = credentials
	
	def build_context(self, *args, **kwargs):
		if self.subprotocol.type == asyauthSubProtocol.NATIVE:
			from asyauth.protocols.spnego.client.native import SPNEGOClientNative
			sspi_ctx = SPNEGOClientNative(self)
			for credential in self.credentials:
				context = credential.build_context()
				if credential.protocol == asyauthProtocol.KERBEROS:
					sspi_ctx.add_auth_context('MS KRB5 - Microsoft Kerberos 5', context)
				elif credential.protocol == asyauthProtocol.NTLM:
					sspi_ctx.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', context)
				else:
					raise Exception('Authentication protocol "%s" is not supported for SPNEGO' % credential.protocol)
			return sspi_ctx

		else:
			raise Exception('Unsupported subprotocol "%s"' % self.subprotocol)


from asyauth.common.constants import asyauthSubProtocol
from asyauth.common.subprotocols import SubProtocol

class SubProtocolWSNetDirect(SubProtocol):
	def __init__(self, ip:str, port:int, proto:str = 'ws'):
		SubProtocol.__init__(self, asyauthSubProtocol.WSNETDIRECT)
		self.ip = ip
		self.port = port
		self.proto = proto

	def get_url(self):
		return "%s://%s:%s" % (self.proto, self.ip, self.port)
	
	@staticmethod
	def from_url_params(query_params = None):
		if 'wsip' not in query_params:
			raise Exception("Query parameter 'wsip' missing")
		ip = query_params['wsip'][0]
		port = query_params.get('wsport', [8700])[0]
		proto = query_params.get('wsproto', ['ws'])[0]
		return SubProtocolWSNetDirect(ip, port, proto)



def hexdump( src, length=16, sep='.', start = 0):
	'''
	@brief Return {src} in hex dump.
	@param[in] length	{Int} Nb Bytes by row.
	@param[in] sep		{Char} For the text part, {sep} will be used for non ASCII char.
	@return {Str} The hexdump
	@note Full support for python2 and python3 !
	'''
	result = []
	if src is None:
		return ''

	# Python3 support
	try:
		xrange(0,1)
	except NameError:
		xrange = range

	for i in xrange(0, len(src), length):
		subSrc = src[i:i+length]
		hexa = ''
		isMiddle = False
		for h in xrange(0,len(subSrc)):
			if h == length/2:
				hexa += ' '
			h = subSrc[h]
			if not isinstance(h, int):
				h = ord(h)
			h = hex(h).replace('0x','')
			if len(h) == 1:
				h = '0'+h
			hexa += h+' '
		hexa = hexa.strip(' ')
		text = ''
		for c in subSrc:
			if not isinstance(c, int):
				c = ord(c)
			if 0x20 <= c < 0x7F:
				text += chr(c)
			else:
				text += sep
		if start == 0:
			result.append(('%08x:  %-'+str(length*(2+1)+1)+'s  |%s|') % (i, hexa, text))
		else:
			result.append(('%08x(+%04x):  %-'+str(length*(2+1)+1)+'s  |%s|') % (start+i, i, hexa, text))
	return '\n'.join(result)


import enum
import io
import os

from asn1crypto.core import ObjectIdentifier

from minikerberos.protocol.constants import EncryptionType
from minikerberos.protocol import encryption
from unicrypto import hmac
from unicrypto.hashlib import md5
from unicrypto.symmetric import RC4

#TODO: RC4 support!

# https://tools.ietf.org/html/draft-raeburn-krb-rijndael-krb-05
# https://tools.ietf.org/html/rfc2478
# https://tools.ietf.org/html/draft-ietf-krb-wg-gssapi-cfx-02
# https://tools.ietf.org/html/rfc4757
# https://www.rfc-editor.org/errata/rfc4757

GSS_WRAP_HEADER     = b'\x60\x2b\x06\x09\x2a\x86\x48\x86\xf7\x12\x01\x02\x02'
GSS_WRAP_HEADER_OID = b'\x60\x2b\x06\x09\x2a\x86\x48\x86\xf7\x12\x01\x02\x02'

class KRB5_MECH_INDEP_TOKEN:
	# https://tools.ietf.org/html/rfc2743#page-81
	# Mechanism-Independent Token Format

	def __init__(self, data, oid, remlen = None):
		self.oid = oid
		self.data = data

		#dont set this
		self.length = remlen
	
	@staticmethod
	def from_bytes(data):
		return KRB5_MECH_INDEP_TOKEN.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		
		start = buff.read(1)
		if start != b'\x60':
			raise Exception('Incorrect token data!')
		remaining_length = KRB5_MECH_INDEP_TOKEN.decode_length_buffer(buff)
		token_data = buff.read(remaining_length)
		
		buff = io.BytesIO(token_data)
		pos = buff.tell()
		buff.read(1)
		oid_length = KRB5_MECH_INDEP_TOKEN.decode_length_buffer(buff)
		buff.seek(pos)
		token_oid = ObjectIdentifier.load(buff.read(oid_length+2))
		
		return KRB5_MECH_INDEP_TOKEN(buff.read(), str(token_oid), remlen = remaining_length)
		
	@staticmethod
	def decode_length_buffer(buff):
		lf = buff.read(1)[0]
		if lf <= 127:
			length = lf
		else:
			bcount = lf - 128
			length = int.from_bytes(buff.read(bcount), byteorder = 'big', signed = False)
		return length
		
	@staticmethod
	def encode_length(length):
		if length <= 127:
			return length.to_bytes(1, byteorder = 'big', signed = False)
		else:
			lb = length.to_bytes((length.bit_length() + 7) // 8, 'big')
			return (128+len(lb)).to_bytes(1, byteorder = 'big', signed = False) + lb
		
		
	def to_bytes(self):
		t = ObjectIdentifier(self.oid).dump() + self.data
		t = b'\x60' + KRB5_MECH_INDEP_TOKEN.encode_length(len(t)) + t
		return t[:-len(self.data)] , self.data


class GSSAPIFlags(enum.IntFlag):
	GSS_C_DCE_STYLE     = 0x1000
	GSS_C_DELEG_FLAG    = 1
	GSS_C_MUTUAL_FLAG   = 2
	GSS_C_REPLAY_FLAG   = 4
	GSS_C_SEQUENCE_FLAG = 8
	GSS_C_CONF_FLAG     = 0x10
	GSS_C_INTEG_FLAG    = 0x20
	
class KG_USAGE(enum.Enum):
	ACCEPTOR_SEAL  = 22
	ACCEPTOR_SIGN  = 23
	INITIATOR_SEAL = 24
	INITIATOR_SIGN = 25
	
class FlagsField(enum.IntFlag):
	SentByAcceptor = 0
	Sealed = 2
	AcceptorSubkey = 4

# https://tools.ietf.org/html/rfc4757 (7.2)
class GSSMIC_RC4:
	def __init__(self):
		self.TOK_ID = b'\x01\x01'
		self.SGN_ALG = b'\x11\x00' #HMAC
		self.Filler = b'\xff'*4
		self.SND_SEQ = None
		self.SGN_CKSUM = None

	@staticmethod
	def from_bytes(data):
		return GSSMIC_RC4.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		mic = GSSMIC_RC4()
		mic.TOK_ID = buff.read(2)
		mic.SGN_ALG = buff.read(2)
		mic.Filler =  buff.read(4)
		mic.SND_SEQ = buff.read(8)
		mic.SGN_CKSUM = buff.read(8)
		
		return mic
		
	def to_bytes(self):
		t  = self.TOK_ID
		t += self.SGN_ALG
		t += self.Filler
		t += self.SND_SEQ
		if self.SGN_CKSUM is not None:
			t += self.SGN_CKSUM
	
		return t
		
class GSSWRAP_RC4:
	def __init__(self):
		self.TOK_ID = b'\x02\x01'
		self.SGN_ALG = b'\x11\x00' #HMAC
		self.SEAL_ALG = None
		self.Filler = b'\xFF' * 2
		self.SND_SEQ = None
		self.SGN_CKSUM = None
		self.Confounder = None

	def __str__(self):
		t  = 'GSSWRAP_RC4\r\n'
		t += 'TOK_ID : %s\r\n' % self.TOK_ID.hex()
		t += 'SGN_ALG : %s\r\n' % self.SGN_ALG.hex()
		t += 'SEAL_ALG : %s\r\n' % self.SEAL_ALG.hex()
		t += 'Filler : %s\r\n' % self.Filler.hex()
		t += 'SND_SEQ : %s\r\n' % self.SND_SEQ.hex()
		t += 'SGN_CKSUM : %s\r\n' % self.SGN_CKSUM.hex()
		t += 'Confounder : %s\r\n' % self.Confounder.hex()
		return t
	
	@staticmethod
	def from_bytes(data):
		return GSSWRAP_RC4.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		wrap = GSSWRAP_RC4()
		wrap.TOK_ID = buff.read(2)
		wrap.SGN_ALG = buff.read(2)
		wrap.SEAL_ALG = buff.read(2)
		wrap.Filler = buff.read(2)
		wrap.SND_SEQ = buff.read(8)
		wrap.SGN_CKSUM = buff.read(8)
		wrap.Confounder = buff.read(8)
		
		return wrap
	
	def to_bytes(self):
		t  = self.TOK_ID
		t += self.SGN_ALG
		t += self.SEAL_ALG
		t += self.Filler
		t += self.SND_SEQ
		
		if self.SGN_CKSUM:
			t += self.SGN_CKSUM
			if self.Confounder:
				t += self.Confounder
		
	
		return t
		
class GSSAPI_RC4:
	def __init__(self, session_key):
		self.session_key = session_key
	
	def GSS_GetMIC(self, data, sequenceNumber, direction = 'init'):
		raise Exception('Not tested! Sure it needs some changes')
		GSS_GETMIC_HEADER = b'\x60\x23\x06\x09\x2a\x86\x48\x86\xf7\x12\x01\x02\x02'
		
		# Let's pad the data
		pad = (4 - (len(data) % 4)) & 0x3
		padStr = bytes([pad]) * pad
		data += padStr
		
		mic = GSSMIC_RC4()
		
		if direction == 'init':
			mic.SND_SEQ = sequenceNumber.to_bytes(4, 'big', signed = False) + b'\x00'*4
		else:
			mic.SND_SEQ = sequenceNumber.to_bytes(4, 'big', signed = False) + b'\xff'*4
		
		Ksign_ctx = hmac_md5(self.session_key.contents, digestmod='md5')
		Ksign_ctx.update(b'signaturekey\0')
		Ksign = Ksign_ctx.digest()
		
		id = 15
		temp = md5( id.to_bytes(4, 'little', signed = False) +  mic.to_bytes()[:8] ).digest()
		chksum_ctx = hmac_md5(Ksign, digestmod='md5')
		chksum_ctx.update(temp)
		mic.SGN_CKSUM = chksum_ctx.digest()[:8]
		
		id = 0
		temp = hmac_md5(self.session_key.contents, digestmod='md5')
		temp.update(id.to_bytes(4, 'little', signed = False))
		
		Kseq_ctx = hmac_md5(temp.digest(), digestmod='md5')
		Kseq_ctx.update(mic.SGN_CKSUM)
		Kseq = Kseq_ctx.digest()
		
		mic.SGN_CKSUM = RC4(Kseq).encrypt(mic.SND_SEQ)
		
		return GSS_GETMIC_HEADER + mic.to_bytes()
		
	
	def GSS_Wrap(self, data, seq_num, direction = 'init', encrypt=True, cofounder = None):
		#direction = 'a'
		#seq_num = 0
		#print('[GSS_Wrap] data: %s' % data)
		#print('[GSS_Wrap] seq_num: %s' % seq_num.to_bytes(4, 'big', signed = False).hex())
		#print('[GSS_Wrap] direction: %s' % direction)
		#print('[GSS_Wrap] encrypt: %s' % encrypt)
		#
		#print('[GSS_Wrap] auth_data: %s' % auth_data)
		
		#pad = 0
		if encrypt is True:
			data += b'\x01'
			#pad = (8 - (len(data) % 8)) & 0x7
			#padStr = bytes([pad]) * pad
			#data += padStr
			#
			##data += b'\x08' * 8
			#print('[GSS_Wrap] pad: %s' % pad)
			#print('[GSS_Wrap] data padded: %s' % data)
		

		token = GSSWRAP_RC4()
		token.SEAL_ALG = b'\x10\x00' # RC4
		
		if direction == 'init':
			token.SND_SEQ = seq_num.to_bytes(4, 'big', signed = False) + b'\x00'*4
		else:
			token.SND_SEQ = seq_num.to_bytes(4, 'big', signed = False) + b'\xff'*4
		
		token.Confounder = os.urandom(8)
		#if cofounder is not None:
		#	token.Confounder = cofounder
		#	#testing purposes only, pls remove
			
		
		temp = hmac.new(self.session_key.contents, digestmod='md5')
		temp.update(b'signaturekey\0')
		Ksign = temp.digest()
		
		id = 13
		Sgn_Cksum = md5(id.to_bytes(4, 'little', signed = False) +  token.to_bytes()[:8] + token.Confounder + data).digest()
		
		klocal = b''
		for b in self.session_key.contents:
			klocal += bytes([b ^ 0xf0])

		id = 0
		temp = hmac.new(klocal, digestmod='md5')
		temp.update(id.to_bytes(4, 'little', signed = False))
		temp = hmac.new(temp.digest(), digestmod='md5')
		temp.update(seq_num.to_bytes(4, 'big', signed = False))
		Kcrypt = temp.digest()

		temp = hmac.new(Ksign, digestmod='md5')
		temp.update(Sgn_Cksum)
		token.SGN_CKSUM = temp.digest()[:8]
		
		id = 0
		temp = hmac.new(self.session_key.contents, digestmod='md5')
		temp.update(id.to_bytes(4, 'little', signed = False))
		temp = hmac.new(temp.digest(), digestmod='md5')
		temp.update(token.SGN_CKSUM)
		Kseq = temp.digest()
		
		token.SND_SEQ = RC4(Kseq).encrypt(token.SND_SEQ)
		
		
		#if auth_data is not None:
		if encrypt is False:
			#print('Unwrap sessionkey: %s' % self.session_key.contents.hex())
			#print('Unwrap data      : %s' % data.hex())

			sspi_wrap = KRB5_MECH_INDEP_TOKEN.from_bytes(data)

			hdr = sspi_wrap.data[:32]
			data = sspi_wrap.data[32:]

			wrap = GSSWRAP_RC4.from_bytes(hdr)
			
			id = 0
			temp = hmac.new(self.session_key.contents, digestmod='md5')
			temp.update(id.to_bytes(4, 'little', signed = False))
			temp = hmac.new(temp.digest(), digestmod='md5')
			temp.update(wrap.SGN_CKSUM)
			Kseq = temp.digest()
			
			snd_seq = RC4(Kseq).encrypt(wrap.SND_SEQ)
			
			id = 0
			temp = hmac.new(klocal, digestmod='md5')
			temp.update(id.to_bytes(4, 'little', signed = False))
			temp = hmac.new(temp.digest(), digestmod='md5')
			temp.update(snd_seq[:4])
			Kcrypt = temp.digest()
			
			rc4 = RC4(Kcrypt)
			dec_cofounder =  rc4.decrypt(wrap.Confounder)
			dec_data = rc4.decrypt(data)

			id = 13
			Sgn_Cksum_calc = md5(id.to_bytes(4, 'little', signed = False) +  wrap.to_bytes()[:8] + dec_cofounder + dec_data).digest()

			temp = hmac.new(Ksign, digestmod='md5')
			temp.update(Sgn_Cksum_calc)
			Sgn_Cksum_calc = temp.digest()[:8]

			if wrap.SGN_CKSUM != Sgn_Cksum_calc[:8]:
				return None, Exception('Integrity verification failed')

			pad = 1
			return dec_data[:-pad], None
			
		elif encrypt is True:
			rc4 = RC4(Kcrypt)
			token.Confounder = rc4.encrypt(token.Confounder)
			cipherText = rc4.encrypt(data)
			finalData, cipherText = KRB5_MECH_INDEP_TOKEN( token.to_bytes() + cipherText,  '1.2.840.113554.1.2.2' ).to_bytes()


			#print('cipherText  %s' % cipherText.hex())
			#print('finalData   %s' % finalData.hex())
			#print('sessionkey  %s' % self.session_key.contents.hex())
			return cipherText, finalData
		

	def GSS_Unwrap(self, data, seq_num, direction='init'):
		#print('GSS_Unwrap data : %s' % data)
		dec_data, err = self.GSS_Wrap(data, seq_num, direction=direction, encrypt = False)
		#print('GSS_Unwrap decrypted data : %s' % dec_data)
		return dec_data, err
	
# 4.2.6.1. MIC Tokens
class GSSMIC:
	def __init__(self):
		self.TOK_ID = b'\x04\x04'
		self.Flags = None
		self.Filler = b'\xFF' * 5
		self.SND_SEQ = None
		self.SGN_CKSUM = None
		
	@staticmethod
	def from_bytes(data):
		return GSSMIC.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		m = GSSMIC()
		m.TOK_ID = buff.read(2)
		m.Flags = FlagsField(int.from_bytes(buff.read(1), 'big', signed = False))
		m.Filler = buff.read(5)
		m.SND_SEQ = int.from_bytes(buff.read(8), 'big', signed = False)
		m.SGN_CKSUM = buff.read() #should know the size based on the algo!
		return m
		
	def to_bytes(self):
		t  = self.TOK_ID
		t += self.Flags.to_bytes(1, 'big', signed = False)
		t += self.Filler
		t += self.SND_SEQ.to_bytes(8, 'big', signed = False)
		if self.SGN_CKSUM is not None:
			t += self.SGN_CKSUM
		
		return t
		
# 4.2.6.2. Wrap Tokens
class GSSWrapToken:
	def __init__(self):
		self.TOK_ID = b'\x05\x04'
		self.Flags = None
		self.Filler = b'\xFF'
		self.EC = None
		self.RRC = None
		self.SND_SEQ = None
		self.Data = None
		
	@staticmethod
	def from_bytes(data):
		return GSSWrapToken.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		m = GSSWrapToken()
		m.TOK_ID = buff.read(2)
		m.Flags = FlagsField(int.from_bytes(buff.read(1), 'big', signed = False))
		m.Filler = buff.read(1)
		m.EC = int.from_bytes(buff.read(2), 'big', signed = False)
		m.RRC = int.from_bytes(buff.read(2), 'big', signed = False)
		m.SND_SEQ = int.from_bytes(buff.read(8), 'big', signed = False)
		return m
		
	def to_bytes(self):
		t  = self.TOK_ID
		t += self.Flags.to_bytes(1, 'big', signed = False)
		t += self.Filler
		t += self.EC.to_bytes(2, 'big', signed = False)
		t += self.RRC.to_bytes(2, 'big', signed = False)
		t += self.SND_SEQ.to_bytes(8, 'big', signed = False)
		if self.Data is not None:
			t += self.Data
		
		return t
	
	def __str__(self):
		t  = 'GSSWrapToken\r\n'
		t += 'TOK_ID : %s\r\n' % self.TOK_ID.hex()
		t += 'Flags : %s\r\n' % self.Flags
		t += 'Filler : %s\r\n' % self.Filler.hex()
		t += 'EC : %s\r\n' % self.EC
		t += 'RRC : %s\r\n' % self.RRC
		t += 'SND_SEQ : %s\r\n' % self.SND_SEQ
		if self.Data is not None:
			t += 'Data : %s\r\n' % self.Data.hex()
		else:
			t += 'Data : None\r\n'
		return t
		
class GSSAPI_AES:
	def __init__(self, session_key, cipher_type, checksum_profile):
		self.session_key = session_key
		self.checksum_profile = checksum_profile
		self.cipher_type = cipher_type
		self.cipher = None
		
	def rotate(self, data, numBytes):
		numBytes %= len(data)
		left = len(data) - numBytes
		result = data[left:] + data[:left]
		return result
		
	def unrotate(self, data, numBytes):
		numBytes %= len(data)
		result = data[numBytes:] + data[:numBytes]
		return result
		
	def GSS_GetMIC(self, data, seq_num):
		pad = (4 - (len(data) % 4)) & 0x3
		padStr = bytes([pad]) * pad
		data += padStr
		
		m = GSSMIC()
		m.Flags = FlagsField.AcceptorSubkey
		m.SND_SEQ = seq_num
		checksum_profile = self.checksum_profile()
		m.checksum = checksum_profile.checksum(self.session_key, KG_USAGE.INITIATOR_SIGN.value, data + m.to_bytes()[:16])
		
		return m.to_bytes()
		
	def GSS_Wrap(self, data, seq_num, use_padding = False):
		#print('[GSS_Wrap] seq_num: %s' % seq_num.to_bytes(4, 'big', signed = False).hex())
		cipher = self.cipher_type()
		pad = 0
		if use_padding is True:
			pad = ((cipher.blocksize - len(data)) % cipher.blocksize) #(cipher.blocksize - (len(data) % cipher.blocksize)) & 15
			padStr = b'\xFF' * pad
			data += padStr
		
		t = GSSWrapToken()
		t.Flags = FlagsField.AcceptorSubkey | FlagsField.Sealed
		t.EC = pad
		t.RRC = 0
		t.SND_SEQ = seq_num
		
		#print('Wrap data: %s' % (data + t.to_bytes()))
		cipher_text = cipher.encrypt(self.session_key, KG_USAGE.INITIATOR_SEAL.value,  data + t.to_bytes(), None)
		t.RRC = 28 #[RFC4121] section 4.2.5
		cipher_text = self.rotate(cipher_text, t.RRC + t.EC)
		
		ret1 = cipher_text
		ret2 = t.to_bytes()

		return ret1, ret2
		
	def GSS_Unwrap(self, data, seq_num, direction='init', auth_data = None, use_padding = False):
		#print('')
		#print('Unwrap data %s' % data[16:])
		#print('Unwrap hdr  %s' % data[:16])

		cipher = self.cipher_type()
		original_hdr = GSSWrapToken.from_bytes(data[:16])
		rotated = data[16:]
		
		cipher_text = self.unrotate(rotated, original_hdr.RRC + original_hdr.EC)
		plain_text = cipher.decrypt(self.session_key, KG_USAGE.ACCEPTOR_SEAL.value, cipher_text)
		new_hdr = GSSWrapToken.from_bytes(plain_text[-16:])

		#signature checking
		new_hdr.RRC = 28
		if data[:16] != new_hdr.to_bytes():
			return None, Exception('GSS_Unwrap signature mismatch!')
		

		#print('Unwrap checksum: %s' % plain_text[-(original_hdr.EC + 16):])
		#print('Unwrap orig chk: %s' % original_hdr.to_bytes())
		#print('Unwrap result 1: %s' % plain_text)
		#print('Unwrap result  : %s' % plain_text[:-(original_hdr.EC + 16)])
		return plain_text[:-(original_hdr.EC + 16)], None
		
def get_gssapi(session_key):
	if session_key.enctype == encryption.Enctype.AES256:
		return GSSAPI_AES(session_key, encryption._AES256CTS, encryption._SHA1AES256)
	if session_key.enctype == encryption.Enctype.AES128:
		return GSSAPI_AES(session_key, encryption._AES128CTS, encryption._SHA1AES128)
	elif session_key.enctype == encryption.Enctype.RC4:
		return GSSAPI_RC4(session_key)
	else:
		raise Exception('Unsupported etype %s' % session_key.enctype)
		
		
def test():
	data = b'\xAF' * 1024
	session_key = encryption.Key( encryption.Enctype.AES256 , bytes.fromhex('3e242e91996aadd513ecb1bc2369e44183e08e08c51550fa4b681e77f75ed8e1'))
	sequenceNumber = 0
	gssapi = get_gssapi(session_key)

	r1, r2 = gssapi.GSS_Wrap(data, sequenceNumber)
	print(len(r2))
	sent = r2 + r1
	print(r1)
	ret1, ret2 = gssapi.GSS_Unwrap(sent, sequenceNumber)

	print(r1.hex())
	print(ret1.hex())


if __name__ == '__main__':
	test()


import os
import io
import base64

from asyauth.protocols.ntlm.structures.fields import Fields
from asyauth.protocols.ntlm.structures.negotiate_flags import NegotiateFlags
from asyauth.protocols.ntlm.structures.version import Version
from asyauth.protocols.ntlm.structures.avpair import AVPairs

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/801a4681-8809-4be9-ab0d-61dcfe762786
class NTLMChallenge:
	def __init__(self):
		self.Signature         = b'NTLMSSP\x00'
		self.MessageType       = 2
		self.TargetNameFields  = None
		self.NegotiateFlags    = None
		self.ServerChallenge   = None
		self.Reserved          = b'\x00'*8
		self.TargetInfoFields  = None
		self.Version           = None
		self.Payload           = None

		self.TargetName        = None
		self.TargetInfo        = None
		
		
	@staticmethod
	def from_bytes(bbuff):
		return NTLMChallenge.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		t = NTLMChallenge()
		t.Signature         = buff.read(8)
		t.MessageType       = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		t.TargetNameFields  = Fields.from_buffer(buff)
		t.NegotiateFlags    = NegotiateFlags(int.from_bytes(buff.read(4), byteorder = 'little', signed = False))
		t.ServerChallenge   = buff.read(8)
		t.Reserved          = buff.read(8)
		t.TargetInfoFields  = Fields.from_buffer(buff)
		
		if t.NegotiateFlags & NegotiateFlags.NEGOTIATE_VERSION: 
			t.Version = Version.from_buffer(buff)
			
		currPos = buff.tell()
		t.Payload = buff.read()
			
		if t.TargetNameFields.length != 0:
			buff.seek(t.TargetNameFields.offset, io.SEEK_SET)
			raw_data = buff.read(t.TargetNameFields.length)
			try:
				t.TargetName = raw_data.decode('utf-16le')
			except UnicodeDecodeError:
				# yet another cool bug. 
				t.TargetName = raw_data.decode('utf-8')
				
		if t.TargetInfoFields.length != 0:
			buff.seek(t.TargetInfoFields.offset, io.SEEK_SET)
			raw_data = buff.read(t.TargetInfoFields.length)
			t.TargetInfo = AVPairs.from_bytes(raw_data)
			
		
		
		return t
	
	@staticmethod
	def construct(challenge = None, targetName = None, targetInfo = None, version = None, flags = None):
		"""
		Construct a new NTLMChallenge message.
		
		Args:
			challenge: 8-byte server challenge (defaults to random)
			targetName: Target name string (optional)
			targetInfo: AVPairs object for target info (optional) 
			version: Version object (optional)
			flags: NegotiateFlags object (required if version is provided)
		
		Returns:
			NTLMChallenge: Constructed challenge message
		"""
		# Generate random challenge if not provided
		if challenge is None:
			challenge = os.urandom(8)
		
		# Validate challenge length
		if len(challenge) != 8:
			raise ValueError("Server challenge must be exactly 8 bytes")
		
		# Create new instance
		t = NTLMChallenge()
		t.ServerChallenge = challenge
		t.Version = version
		
		# Handle flags - set NEGOTIATE_VERSION if version is provided
		if flags is None:
			flags = NegotiateFlags(0)
		elif not isinstance(flags, NegotiateFlags):
			flags = NegotiateFlags(flags)
		
		if version is not None:
			flags |= NegotiateFlags.NEGOTIATE_VERSION
		
		t.NegotiateFlags = flags
		
		# Calculate payload position after fixed fields
		# Fixed header: 8 (sig) + 4 (type) + 8 (targetname fields) + 4 (flags) + 8 (challenge) + 8 (reserved) + 8 (targetinfo fields) = 48
		payload_pos = 48
		if version is not None:
			payload_pos += 8  # Version structure is 8 bytes
		
		# Handle target name
		if targetName is not None:
			target_name_data = targetName.encode('utf-16le')
			t.TargetName = targetName
			t.TargetNameFields = Fields(len(target_name_data), payload_pos)
			payload_pos += len(target_name_data)
		else:
			t.TargetName = ''
			t.TargetNameFields = Fields(0, payload_pos)
		
		# Handle target info
		if targetInfo is not None:
			target_info_data = targetInfo.to_bytes()
			t.TargetInfo = targetInfo
			t.TargetInfoFields = Fields(len(target_info_data), payload_pos)
		else:
			# Create empty AVPairs with just EOL marker
			t.TargetInfo = AVPairs()
			target_info_data = t.TargetInfo.to_bytes()
			t.TargetInfoFields = Fields(len(target_info_data), payload_pos)
		
		# Build payload
		t.Payload = b''
		if targetName is not None:
			t.Payload += targetName.encode('utf-16le')
		if targetInfo is not None:
			t.Payload += targetInfo.to_bytes()
		else:
			t.Payload += t.TargetInfo.to_bytes()
		
		return t

	def to_bytes(self):
		tn = self.TargetName.encode('utf-16le')
		ti = self.TargetInfo.to_bytes()

		buff  = self.Signature
		buff += self.MessageType.to_bytes(4, byteorder = 'little', signed = False)
		buff += self.TargetNameFields.to_bytes()
		buff += self.NegotiateFlags.to_bytes(4, byteorder = 'little', signed = False)
		buff += self.ServerChallenge
		buff += self.Reserved
		buff += self.TargetInfoFields.to_bytes()
		if self.Version:
			buff += self.Version.to_bytes()
		buff += self.Payload

		return buff

	def __repr__(self):
		t  = '== NTLMChallenge ==\r\n'
		t += 'Signature      : %s\r\n' % repr(self.Signature)
		t += 'MessageType    : %s\r\n' % repr(self.MessageType)
		t += 'ServerChallenge: %s\r\n' % repr(self.ServerChallenge)
		t += 'TargetName     : %s\r\n' % repr(self.TargetName)
		t += 'TargetInfo     : %s\r\n' % repr(self.TargetInfo)
		return t

	def toBase64(self):
		return base64.b64encode(self.to_bytes()).decode('ascii')


def test():
	test_reconstrut()
	test_construct()
	test_template()
	
def test_reconstrut(data = None):
	try:
		from asyauth.utils.hexdump import hexdump
	except ImportError:
		def hexdump(data):
			return data.hex()
	
	print('=== reconstruct===')
	if not data:
		challenge_test_data = bytes.fromhex('4e544c4d53535000020000000800080038000000158289e2a7314a557bdb11bf000000000000000072007200400000000a0063450000000f540045005300540002000800540045005300540001001200570049004e003200300031003900410044000400120074006500730074002e0063006f007200700003002600570049004e003200300031003900410044002e0074006500730074002e0063006f007200700007000800aec600bfc5fdd40100000000')
	else:
		challenge_test_data = data
	challenge = NTLMChallenge.from_bytes(challenge_test_data)
	print(repr(challenge))
	challenge_test_data_verify = challenge.to_bytes()
	print('====== reconstructed ====')
	print(hexdump(challenge_test_data_verify))
	print('====== original ====')
	print(hexdump(challenge_test_data))
	assert challenge_test_data == challenge_test_data_verify
	
def test_template():
	
	#challenge = NTLMChallenge.construct_from_template('Windows2003')
	#test_reconstrut(challenge.to_bytes())
	pass
	
def test_construct():
	print('=== test_construct ===')
	
	# Test basic construction with minimal parameters
	challenge1 = NTLMChallenge.construct()
	print('Basic construct test passed')
	
	# Test with custom challenge
	custom_challenge = b'\x01\x02\x03\x04\x05\x06\x07\x08'
	challenge2 = NTLMChallenge.construct(challenge=custom_challenge)
	assert challenge2.ServerChallenge == custom_challenge
	print('Custom challenge test passed')
	
	# Test with target name
	challenge3 = NTLMChallenge.construct(targetName="DOMAIN")
	assert challenge3.TargetName == "DOMAIN"
	print('Target name test passed')
	
	# Test with target info
	from asyauth.protocols.ntlm.structures.avpair import AVPairs, AVPAIRType
	target_info = AVPairs()
	target_info[AVPAIRType.MsvAvNbDomainName] = "DOMAIN"
	challenge4 = NTLMChallenge.construct(targetInfo=target_info)
	assert challenge4.TargetInfo[AVPAIRType.MsvAvNbDomainName] == "DOMAIN"
	print('Target info test passed')
	
	# Test roundtrip (construct -> to_bytes -> from_bytes)
	challenge5 = NTLMChallenge.construct(
		challenge=custom_challenge,
		targetName="TEST",
		targetInfo=target_info
	)
	data = challenge5.to_bytes()
	challenge6 = NTLMChallenge.from_bytes(data)
	assert challenge6.ServerChallenge == custom_challenge
	assert challenge6.TargetName == "TEST"
	print('Roundtrip test passed')
	
	print('All construct tests passed!')
	
if __name__ == '__main__':
	from asyauth.utils.hexdump import hexdump
	
	test()


import copy
import os
import struct
from typing import cast, Callable
from asyauth.common.winapi.constants import ISC_REQ
from asyauth.common.constants import asyauthSecret, asyauthProtocol, asyauthSubProtocol


from unicrypto import hmac
from unicrypto import hashlib
from unicrypto.symmetric import RC4

from asyauth.protocols.ntlm import logger
from asyauth.common.credentials.ntlm import NTLMCredential
from asyauth.protocols.ntlm.structures.serverinfo import NTLMServerInfo
from asyauth.protocols.ntlm.structures.negotiate_flags import NegotiateFlags
from asyauth.protocols.ntlm.structures.ntlmssp_message_signature import NTLMSSP_MESSAGE_SIGNATURE
from asyauth.protocols.ntlm.structures.ntlmssp_message_signature_noext import NTLMSSP_MESSAGE_SIGNATURE_NOEXT
from asyauth.protocols.ntlm.messages.negotiate import NTLMNegotiate
from asyauth.protocols.ntlm.messages.challenge import NTLMChallenge
from asyauth.protocols.ntlm.messages.authenticate import NTLMAuthenticate
from asyauth.protocols.ntlm.creds_calc import netntlmv2, AVPAIRType, LMResponse, netntlm, netntlm_ess
from minikerberos.gssapi.channelbindings import ChannelBindingsStruct
from asyauth.protocols.ntlm.structures.avpair import AVPairs, MsvAvFlags
from asyauth.protocols.ntlm.structures.version import Version
from asyauth.protocols.ntlm.structures.serverinfo import FILETIME
from datetime import datetime
		

class NTLMServerNative:
	def __init__(self, credential:NTLMCredential):
		self.credential = credential
		self.challenge = credential.challenge if credential.challenge is not None else os.urandom(8)
		
		self.ntlmNegotiate     = None #ntlm Negotiate message from client
		self.ntlmChallenge     = None #ntlm Challenge message to client
		self.ntlmAuthenticate  = None #ntlm Authenticate message from client
		
		self.ntlmNegotiate_raw     = None #message as bytes, as it's recieved/sent
		self.ntlmChallenge_raw     = None #message as bytes, as it's recieved/sent
		self.ntlmAuthenticate_raw  = None #message as bytes, as it's recieved/sent

		
		self.EncryptedRandomSessionKey = None
		self.RandomSessionKey = credential.session_key if credential.session_key is not None else os.urandom(16)
		self.SessionBaseKey = None
		self.KeyExchangeKey = None
		
		self.SignKey_client = None
		self.SealKey_client = None
		self.SignKey_server = None
		self.SealKey_server = None

		self.crypthandle_client = None
		self.crypthandle_server = None
		
		self.seq_number = 0
		self.iteration_cnt = 0
		self.ntlm_credentials = None
		self.timestamp = credential.timestamp #used in unittest only!
		self.extra_info = None
			
	def load_negotiate(self, data):
		logger.debug('Loading negotiate message')
		self.ntlmNegotiate = NTLMNegotiate.from_bytes(data)
	
	def load_challenge(self, data):
		logger.debug('Loading challenge message')
		self.ntlmChallenge = NTLMChallenge.from_bytes(data)
		
	def load_authenticate(self, data):
		logger.debug('Loading authenticate message')
		self.ntlmAuthenticate = NTLMAuthenticate.from_bytes(data)
		
	def load_sessionkey(self, data):
		logger.debug('Loading sessionkey')
		self.RandomSessionKey = data
		self.setup_crypto(True)
	
	def is_guest(self):
		return self.credential.is_guest

	def get_seq_number(self):
		return self.seq_number
	
	def set_sign(self, tf = True):
		if tf == True:
			self.credential.flags |= NegotiateFlags.NEGOTIATE_SIGN
		else:
			self.credential.flags &= ~NegotiateFlags.NEGOTIATE_SIGN
			
	def set_seal(self, tf = True):
		if tf == True:
			self.credential.flags |= NegotiateFlags.NEGOTIATE_SEAL
		else:
			self.credential.flags &= ~NegotiateFlags.NEGOTIATE_SEAL
			
	def set_version(self, tf = True):
		if tf == True:
			self.credential.flags |= NegotiateFlags.NEGOTIATE_VERSION
		else:
			self.credential.flags &= ~NegotiateFlags.NEGOTIATE_VERSION
			
	def is_extended_security(self):
		return NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY in self.ntlmChallenge.NegotiateFlags
	
	def get_extra_info(self):
		self.extra_info = NTLMServerInfo.from_challenge(self.ntlmChallenge)
		return self.extra_info
		
	def MAC(self, handle, signingKey, seqNum, message):
		if self.is_extended_security() == True:
			msg = NTLMSSP_MESSAGE_SIGNATURE()
			if NegotiateFlags.NEGOTIATE_KEY_EXCH in self.ntlmChallenge.NegotiateFlags:
				tt = struct.pack('<i', seqNum) + message
				t = hmac.new(signingKey, digestmod='md5')
				t.update(tt)
				
				msg.Checksum = handle(t.digest()[:8])
				msg.SeqNum = seqNum
				seqNum += 1
			else:
				t = hmac.new(signingKey, digestmod='md5')
				t.update(struct.pack('<i',seqNum)+message)
				msg.Checksum = t.digest()[:8]
				msg.SeqNum = seqNum
				seqNum += 1
				
		else:
			raise Exception('Not implemented!')
			#t = struct.pack('<I',binascii.crc32(message)& 0xFFFFFFFF)
			#randompad = 0
			#msg = NTLMSSP_MESSAGE_SIGNATURE_NOEXT()
			#msg.RandomPad = handle(struct.pack('<I',randompad))
			#msg.Checksum = struct.unpack('<I',handle(messageSignature['Checksum']))[0]
			
		return msg.to_bytes()

	async def encrypt(self, data, sequence_no):
		"""
		This function is to support SSPI encryption.
		"""
		data = self.SEAL(
			self.SignKey_client,
			self.SealKey_client, 
			data,
			data,
			sequence_no, 
			self.crypthandle_client.encrypt
		)
		return data

	async def decrypt(self, data, sequence_no, direction='init', auth_data=None):
		"""
		This function is to support SSPI decryption.
		"""
		edata = data[16:]
		srv_sig = NTLMSSP_MESSAGE_SIGNATURE.from_bytes(data[:16])
		sealedMessage = self.crypthandle_server.encrypt(edata)
		signature = self.MAC(self.crypthandle_server.encrypt, self.SignKey_server, srv_sig.SeqNum, sealedMessage)
		#print('seqno     %s' % sequence_no)
		#print('Srv  sig: %s' % data[:16])
		#print('Calc sig: %s' % signature)

		return sealedMessage, None

	async def sign(self, data, message_no, direction=None, reset_cipher = False):
		"""
		Singing outgoing messages. The reset_cipher parameter is needed for calculating mechListMIC. 
		"""
		#print('sign data : %s' % data)
		#print('sign message_no : %s' % message_no)
		#print('sign direction : %s' % direction)
		signature = self.MAC(self.crypthandle_client.encrypt, self.SignKey_client, message_no, data)
		if reset_cipher is True:
			self.crypthandle_client = RC4(self.SealKey_client)
			self.crypthandle_server = RC4(self.SealKey_server)
		self.seq_number += 1
		return signature

	async def verify(self, data, signature):
		"""
		Verifying incoming server message
		"""
		signature_struct = NTLMSSP_MESSAGE_SIGNATURE.from_bytes(signature)
		calc_sig = self.MAC(self.crypthandle_server.encrypt, self.SignKey_server, signature_struct.SeqNum, data)
		#print('server signature    : %s' % signature)
		#print('calculates signature: %s' % calc_sig)
		return signature == calc_sig

	def SEAL(self, signingKey, sealingKey, messageToSign, messageToEncrypt, seqNum, cipher_encrypt):
		"""
		This is the official SEAL function.
		"""
		sealedMessage = cipher_encrypt(messageToEncrypt)
		signature = self.MAC(cipher_encrypt, signingKey, seqNum, messageToSign)
		return sealedMessage, signature
		
	def SIGN(self, signingKey, message, seqNum, cipher_encrypt):
		"""
		This is the official SIGN function.
		"""
		return self.MAC(cipher_encrypt, signingKey, seqNum, message)
	
	def signing_needed(self):
		return (
			NegotiateFlags.NEGOTIATE_SIGN in self.ntlmChallenge.NegotiateFlags or \
			NegotiateFlags.NEGOTIATE_SEAL in self.ntlmChallenge.NegotiateFlags
		)
	
	def encryption_needed(self):
		return NegotiateFlags.NEGOTIATE_SEAL in self.ntlmChallenge.NegotiateFlags

	def calc_sealkey(self, mode = 'Client'):
		if NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY in self.ntlmChallenge.NegotiateFlags:
			if NegotiateFlags.NEGOTIATE_128 in self.ntlmChallenge.NegotiateFlags:
				sealkey = self.RandomSessionKey
			elif NegotiateFlags.NEGOTIATE_56 in self.ntlmChallenge.NegotiateFlags:
				sealkey = self.RandomSessionKey[:7]
			else:
				sealkey = self.RandomSessionKey[:5]
				
			if mode == 'Client':
				md5 = hashlib.new('md5')
				md5.update(sealkey + b'session key to client-to-server sealing key magic constant\x00')
				sealkey = md5.digest()
			else:
				md5 = hashlib.new('md5')
				md5.update(sealkey + b'session key to server-to-client sealing key magic constant\x00')
				sealkey = md5.digest()
				
		elif NegotiateFlags.NEGOTIATE_56 in self.ntlmChallenge.NegotiateFlags:
			sealkey = self.RandomSessionKey[:7] + b'\xa0'
		else:
			sealkey = self.RandomSessionKey[:5] + b'\xe5\x38\xb0'
			
		if mode == 'Client':
			self.SealKey_client = sealkey
			if sealkey is not None:
				self.crypthandle_client = RC4(self.SealKey_client)
		else:
			self.SealKey_server = sealkey
			if sealkey is not None:
				self.crypthandle_server = RC4(self.SealKey_server)
		
		if sealkey is not None:
			logger.debug('Setting %s sealkey to %s' % (mode, sealkey.hex()))
		return sealkey
		
	def calc_signkey(self, mode = 'Client'):
		if NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY in self.ntlmChallenge.NegotiateFlags:
			if mode == 'Client':
				md5 = hashlib.new('md5')
				md5.update(self.RandomSessionKey + b"session key to client-to-server signing key magic constant\x00")
				signkey = md5.digest()
			else:
				md5 = hashlib.new('md5')
				md5.update(self.RandomSessionKey + b"session key to server-to-client signing key magic constant\x00")
				signkey = md5.digest()
		else:
			signkey = None
			
		if mode == 'Client':
			if signkey is not None:
				logger.debug('Setting client signkey to %s' % signkey.hex())
			self.SignKey_client = signkey

		else:
			if signkey is not None:
				logger.debug('Setting server signkey to %s' % signkey.hex())
			self.SignKey_server = signkey
		
		return signkey
		
	def get_session_key(self):
		return self.RandomSessionKey
		
	def get_sealkey(self, mode = 'Client'):
		if mode == 'Client':
			return self.SealKey_client
		else:
			return self.SealKey_server
			
	def get_signkey(self, mode = 'Client'):
		if mode == 'Client':
			return self.SignKey_client
		else:
			return self.SignKey_server
		
	def setup_crypto(self, is_remote = False):
		logger.debug('Setting up crypto')
		if not self.RandomSessionKey:
			self.RandomSessionKey = os.urandom(16)
			logger.debug('RandomSessionKey: %s' % self.RandomSessionKey.hex())
		
		
		if self.credential.is_guest == True:
			self.SessionBaseKey = b'\x00' * 16
			self.KeyExchangeKey = b'\x00' * 16
				
			rc4 = RC4(self.KeyExchangeKey)
			self.EncryptedRandomSessionKey = rc4.encrypt(self.RandomSessionKey)
			logger.debug('EncryptedRandomSessionKey: %s' % self.EncryptedRandomSessionKey.hex())

		else:
			if is_remote is False:
				#this check is here to provide the option to load the messages + the sessionbasekey manually
				#then you will be able to use the sign and seal functions provided by this class
				self.SessionBaseKey = self.ntlm_credentials.SessionBaseKey
			
				rc4 = RC4(self.KeyExchangeKey)
				self.EncryptedRandomSessionKey = rc4.encrypt(self.RandomSessionKey)
				logger.debug('EncryptedRandomSessionKey: %s' % self.EncryptedRandomSessionKey.hex())
		
		self.calc_sealkey('Client')
		self.calc_sealkey('Server')
		self.calc_signkey('Client')
		self.calc_signkey('Server')

	def isc_to_ntlm_flags(self, flags):
		# trying to guess what ISC flags match which
		if isinstance(flags, NegotiateFlags) is True:
			return flags
		if flags is None:
			# using the pre-defined flags
			return self.credential.flags
		ntlmflags = copy.deepcopy(self.credential.flags)
		ntlmflags |= NegotiateFlags.NEGOTIATE_56
		ntlmflags |= NegotiateFlags.NEGOTIATE_KEY_EXCH
		ntlmflags |= NegotiateFlags.NEGOTIATE_128
		ntlmflags |= NegotiateFlags.NEGOTIATE_VERSION
		ntlmflags |= NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY
		ntlmflags |= NegotiateFlags.NEGOTIATE_NTLM
		ntlmflags |= NegotiateFlags.NEGOTIATE_LM_KEY
		ntlmflags |= NegotiateFlags.REQUEST_TARGET
		ntlmflags |= NegotiateFlags.NTLM_NEGOTIATE_OEM
		ntlmflags |= NegotiateFlags.NEGOTIATE_UNICODE
		ntlmflags &= ~NegotiateFlags.TARGET_TYPE_SERVER
		
		if ISC_REQ.INTEGRITY in flags:
			ntlmflags |= NegotiateFlags.NEGOTIATE_SIGN
		else:
			ntlmflags  &= ~NegotiateFlags.NEGOTIATE_SIGN

		
		if ISC_REQ.CONFIDENTIALITY in flags:
			ntlmflags |= NegotiateFlags.NEGOTIATE_SEAL
		else:
			ntlmflags  &= ~NegotiateFlags.NEGOTIATE_SEAL
		
		return ntlmflags

	async def authenticate_server(self, authData, flags = None, cb_data = None, spn=None, **kwargs):
		# check if kwargs have target info
		target = None
		try:
			if 'target' in kwargs:
				from asysocks.unicomm.common.target import UniTarget
				target = cast(UniTarget, kwargs['target'])
		except:
			pass
			
		flags = self.isc_to_ntlm_flags(flags)
		logger.debug('Flags: %s' % flags)
		if self.iteration_cnt == 0:
			if authData is None:
				raise Exception('First call as server MUST be with data!')
				
			self.iteration_cnt += 1
			#negotiate message was already calulcated in setup
			self.ntlmNegotiate = NTLMNegotiate.from_bytes(authData)
			self.ntlmNegotiate_raw = authData

			targetinfo = AVPairs()
			targetinfo[AVPAIRType.MsvAvNbDomainName] = 'TEST'
			targetinfo[AVPAIRType.MsvAvNbComputerName] = 'TEST$'

			targetinfo[AVPAIRType.MsvAvDnsDomainName] = 'TEST.local'
			targetinfo[AVPAIRType.MsvAvDnsComputerName] = 'TEST.local'
			targetinfo[AVPAIRType.MsvAvDnsTreeName] = 'TEST.local'
			targetinfo[AVPAIRType.MsvAvTimestamp] = FILETIME.from_datetime(datetime.now()).to_bytes()
			#targetinfo[AVPAIRType.MsvAvFlags] = MsvAvFlags.MsvAvFlags(0x00000002)
			

			flags = copy.deepcopy(self.ntlmNegotiate.NegotiateFlags)
			flags |= NegotiateFlags.NEGOTIATE_VERSION
			flags |= NegotiateFlags.NEGOTIATE_TARGET_INFO
			flags &= ~NegotiateFlags.NEGOTIATE_ALWAYS_SIGN
			flags &= ~NegotiateFlags.NEGOTIATE_SEAL
			flags &= ~NegotiateFlags.NEGOTIATE_SIGN
			flags &= ~NegotiateFlags.TARGET_TYPE_SERVER

			self.ntlmChallenge = NTLMChallenge.construct(
				challenge = os.urandom(8),
				targetName = self.credential.negotiate_domain if self.credential.negotiate_domain is not None else '',
				targetInfo = targetinfo,
				version = self.credential.negotiate_version,
				flags = flags
			)
			
			logger.debug('Negotiate: %s' % self.ntlmNegotiate)
			self.ntlmChallenge_raw = self.ntlmChallenge.to_bytes()
			return self.ntlmChallenge_raw, True, None
			
		else:
			#client authenticate incoming
			self.ntlmAuthenticate_raw = authData
			self.ntlmAuthenticate = NTLMAuthenticate.from_bytes(authData)
			logger.debug('Authenticate: %s' % self.ntlmAuthenticate)
				
			##################flags = self.ntlmChallenge.NegotiateFlags
				
			#we need to calculate the response based on the credential and the settings flags
			if self.credential.ntlm_version == 1:
				return None, False, None
				#raise NotImplementedError('NTLMv1 is not supported as server!')
				logger.debug('NTLMv1 is used here')
				#NTLMv1 authentication
				# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/464551a8-9fc4-428e-b3d3-bc5bfb2e73a5
					
				#check if we authenticate as guest
				if self.credential.is_guest == True:
					logger.debug('NTLMv1 Guest authentication')
					lmresp = LMResponse()
					self.set_version(False)
					lmresp.Response = b'\x00'
					self.ntlmAuthenticate = NTLMAuthenticate.construct(
						flags,
						lm_response= lmresp,
						mic=None,
						encrypted_session = self.EncryptedRandomSessionKey,
						workstationname=target.hostname if target is not None else None
					)
					logger.debug('NTLMv1 Guest - NTLMAuthenticate: %s' % self.ntlmAuthenticate)
					return self.ntlmAuthenticate.to_bytes(), False, None
						
				if flags & NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY:
					logger.debug('NTLMv1 with extended security')
					self.ntlm_credentials = netntlm_ess.construct(self.ntlmChallenge.ServerChallenge, self.challenge, self.credential)
					
					self.KeyExchangeKey = self.ntlm_credentials.calc_key_exchange_key()
					self.setup_crypto()
						
					self.ntlmAuthenticate = NTLMAuthenticate.construct(
						flags,
						lm_response= self.ntlm_credentials.LMResponse,
						nt_response = self.ntlm_credentials.NTResponse,
						version = self.ntlmNegotiate.Version,
						encrypted_session = self.EncryptedRandomSessionKey,
						workstationname=target.hostname if target is not None else None
					)
					logger.debug('NTLMv1 with extended security - NTLMAuthenticate: %s' % self.ntlmAuthenticate)
				else:
					self.ntlm_credentials = netntlm.construct(self.ntlmChallenge.ServerChallenge, self.credential)
						
					self.KeyExchangeKey = self.ntlm_credentials.calc_key_exchange_key(with_lm = flags & NegotiateFlags.NEGOTIATE_LM_KEY, non_nt_session_key = flags & NegotiateFlags.REQUEST_NON_NT_SESSION_KEY)						
					self.setup_crypto()
					self.ntlmAuthenticate = NTLMAuthenticate.construct(
						flags,
						lm_response= self.ntlm_credentials.LMResponse,
						nt_response = self.ntlm_credentials.NTResponse,
						version = self.ntlmNegotiate.Version,
						encrypted_session = self.EncryptedRandomSessionKey,
						workstationname=target.hostname if target is not None else None
					)
					logger.debug('NTLMv1 - NTLMAuthenticate: %s' % self.ntlmAuthenticate)
							
							
			else:
				#NTLMv2
				# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/5e550938-91d4-459f-b67d-75d70009e3f3
				if self.ntlmAuthenticate.UserName == '':
					# allowing guests here
					return None, False, None
					#lmresp = LMResponse()
					#lmresp.Response = b'\x00'
					#self.set_version(False)
					#self.setup_crypto()
					#self.ntlmAuthenticate = NTLMAuthenticate.construct(
					#	flags,
					#	lm_response= lmresp,
					#	encrypted_session = self.EncryptedRandomSessionKey,
					#	workstationname=target.hostname if target is not None else None
					#)						
					#logger.debug('NTLMv2 Guest - NTLMAuthenticate: %s' % self.ntlmAuthenticate)
					#return self.ntlmAuthenticate.to_bytes(), False, None
						
				else:
					return None, False, None
					#comment this out for testing!
					ti = self.ntlmChallenge.TargetInfo
					if spn is not None:
						ti[AVPAIRType.MsvAvTargetName] = spn
					if cb_data is not None:
						cb_struct = ChannelBindingsStruct()
						cb_struct.application_data = cb_data

						md5_ctx = hashlib.new('md5')
						md5_ctx.update(cb_struct.to_bytes())
						ti[AVPAIRType.MsvChannelBindings] = md5_ctx.digest()
					###
						
					self.ntlm_credentials = netntlmv2.construct(self.ntlmChallenge.ServerChallenge, self.challenge, ti, self.credential, timestamp = self.timestamp)
					self.KeyExchangeKey = self.ntlm_credentials.calc_key_exchange_key()
					logger.debug('NTLMv2 - KeyExchangeKey: %s' % self.KeyExchangeKey.hex())				
					self.setup_crypto()
						
					#TODO: if "ti" / targetinfo in the challenge message has "MsvAvFlags" type and the bit for MIC is set (0x00000002) we need to send a MIC. probably...
					mic = None
						
					self.ntlmAuthenticate = NTLMAuthenticate.construct(
						flags,
						domainname= self.credential.domain,
						workstationname=target.hostname if target is not None else None,
						username= self.credential.username,
						lm_response= self.ntlm_credentials.LMResponse,
						nt_response= self.ntlm_credentials.NTResponse,
						version = self.ntlmNegotiate.Version,
						encrypted_session = self.EncryptedRandomSessionKey,
						mic = mic,
					)
				
					logger.debug('NTLMv2 - NTLMAuthenticate: %s' % self.ntlmAuthenticate)
			self.ntlmAuthenticate_raw = self.ntlmAuthenticate.to_bytes()
			return self.ntlmAuthenticate_raw, False, None
	
def ntlmserver_factory(ntlm_credential_factory: Callable = None) -> NTLMServerNative:
	if ntlm_credential_factory is None:
		ntlm_settings = NTLMCredential(
			'test',
			'test',
			'test',
			stype = asyauthSecret.PASSWORD
		)
	else:
		ntlm_settings = ntlm_credential_factory()
	ntlm_ctx = NTLMServerNative(ntlm_settings)
	return ntlm_ctx


import enum
import io
import uuid
import os

# https://tools.ietf.org/html/draft-zhu-negoex-04

# UCHAR is the data type for a one-octet number.
# ULONG is the data type for a 4-octet number encoded in little endian.
# USHORT is the data type for a 2-octet number encoded in little endian.
# ULONG64 is the data type for a 8-octet number encoded in little endian.
# GUID is the data type for a 16-octet number encoded in little endian.


from minikerberos.protocol.asn1_structs import AS_REQ, AS_REP, AP_REQ, AP_REP
from minikerberos.protocol.rfc4556 import MetaData

NEGOEXTS_DEFAULT_AUTHID = uuid.UUID(bytes_le=bytes.fromhex('5c33530deaf90d4db2ec4ae3786ec308'))

def cl_bytes(data):
	return cl_buff(io.BytesIO(data))

def cl_buff(buff):
	x = buff.read(1)
	if x[0] <= 127:
		return x[0]
	else:
		bcount = x[0] - 128
		y = buff.read(bcount)
		return int.from_bytes(y, byteorder = 'big', signed = False)

class PKU2U_TOKEN_TYPE(enum.Enum):
	KRB_AP_REQ = b'\x01\x00'
	KRB_AP_REP = b'\x02\x00'
	KRB_ERROR  = b'\x03\x00'
	KRB_AS_REQ = b'\x05\x00'
	KRB_AS_REP = b'\x06\x00'

class PKU2U_TOKEN:
	def __init__(self, tok_id = b'\x01\x00'):
		self.tok_id = tok_id
		self.inner_token = None
		#self.inner_token_raw = None #unparsed bytes
		self._pku2u_oid = bytes.fromhex('2b0601050207')
	
	@staticmethod
	def from_bytes(data):
		return PKU2U_TOKEN.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		t = PKU2U_TOKEN()
		t_hdr = buff.tell()
		buff.read(1) # 0x60
		total_length = cl_buff(buff)
		buff.read(1) # 0x06
		total_length += buff.tell() - t_hdr - 1
		oid_length = cl_buff(buff)
		t_oid = buff.read(oid_length)
		t.tok_id = PKU2U_TOKEN_TYPE(buff.read(2))
		t_data = buff.read(total_length - buff.tell())
		#t.inner_token_raw = t_data

		if t.tok_id == PKU2U_TOKEN_TYPE.KRB_AS_REQ:
			t.inner_token = AS_REQ.load(t_data)
		elif t.tok_id == PKU2U_TOKEN_TYPE.KRB_AS_REP:
			t.inner_token = AS_REP.load(t_data)
		elif t.tok_id == PKU2U_TOKEN_TYPE.KRB_AP_REQ:
			t.inner_token = AP_REQ.load(t_data)
		elif t.tok_id == PKU2U_TOKEN_TYPE.KRB_AP_REP:
			t.inner_token = AP_REP.load(t_data)
		else:
			t.inner_token = t_data
		return t
		
	def length_encode(self, x):
		if x <= 127:
			return x.to_bytes(1, 'big', signed = False)
		else:
			lb = x.to_bytes((x.bit_length() + 7) // 8, 'big')
			t = (0x80 | len(lb)).to_bytes(1, 'big', signed = False)
			return t+lb
		
	def to_bytes(self):
		data = b'\x06'
		data += self.length_encode(len(self._pku2u_oid))
		data += self._pku2u_oid
		data += self.tok_id.value
		data += self.inner_token

		t = b'\x60'
		t += self.length_encode(len(data)) #+2 ?
		t += data
		return t


class MESSAGE_TYPE(enum.Enum):
	INITIATOR_NEGO = 0
	ACCEPTOR_NEGO = 1
	INITIATOR_META_DATA = 2
	ACCEPTOR_META_DATA = 3
	CHALLENGE = 4
	AP_REQUEST = 5
	VERIFY = 6
	ALERT = 7

NEGOEXT_MESSAGE_TYPES = [MESSAGE_TYPE.INITIATOR_NEGO, MESSAGE_TYPE.CHALLENGE, MESSAGE_TYPE.INITIATOR_META_DATA, MESSAGE_TYPE.ACCEPTOR_META_DATA, MESSAGE_TYPE.AP_REQUEST,  MESSAGE_TYPE.VERIFY,  MESSAGE_TYPE.ALERT]


class BYTE_VECTOR:
	"""
		BYTE_VECTOR encapsulates a variable length array of octets (or bytes)
		that are stored consecutively.  Each element in is a byte (8 bits).
	"""
	def __init__(self):
		self.ByteArrayOffset = None #ULONG 
		self.ByteArrayLength = None #ULONG

class AUTH_SCHEME_VECTOR:
	"""
		AUTH_SCHEME_VECTOR encapsulates a variable length array of
		AUTH_SCHEMEs that are stored consecutively.  Each element is a
		structure of the type AUTH_SCHEME.
	"""
	def __init__(self):
		self.AuthSchemeArrayOffset = None #ULONG // each element contains an AUTH_SCHEME
		self.AuthSchemeCount = None #ULONG

class EXTENSION_VECTOR:
	"""
		EXTENSION_VECTOR encapsulates a variable length array of EXTENSIONs
		that are stored consecutively.  Each element is a structure of the
		type EXTENSION.

	"""
	def __init__(self):
		self.ExtensionArrayOffset = None #ULONG // each element contains an AUTH_SCHEME
		self.ExtensionCount = None #ULONG

class EXTENSION:
	"""
	This part is not tested, as the RFC doesn't define any extensions at the moment!
	"""
	def __init__(self):
		self.ExtensionType = None #ULONG 
		self.ExtensionArrayOffset = None #ULONG 
		self.ExtensionArrayLength = None #ULONG

		self.ExtensionValue = None #BYTE_VECTOR

	@staticmethod
	def from_bytes(data):
		return EXTENSION.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		msg = EXTENSION()
		msg.ExtensionType = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.ExtensionArrayOffset = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.ExtensionArrayLength = int.from_bytes(buff.read(4), byteorder='little', signed = False)

		buff.seek(msg.ExtensionArrayOffset)
		msg.ExtensionValue = buff.read(msg.ExtensionArrayLength)
		return msg

class CHECKSUM:
	#define CHECKSUM_SCHEME_RFC3961 1 - Currently only one value is defined.
	def __init__(self):
		self.cbHeaderLength = 20 #ULONG, always 20
		self.ChecksumScheme = None #ULONG 
		self.ChecksumType = None #ULONG // in the case of RFC3961 scheme, this is the RFC3961 checksum type
		self.ChecksumArrayOffset = None
		self.ChecksumArrayLength = None
		
		self.ChecksumValue = None #BYTE_VECTOR

	@staticmethod
	def from_bytes(data):
		return CHECKSUM.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		start_offset = buff.tell()
		msg = CHECKSUM()
		msg.cbHeaderLength = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.ChecksumScheme = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.ChecksumType = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.ChecksumArrayOffset = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.ChecksumArrayLength = int.from_bytes(buff.read(4), byteorder='little', signed = False)

		if msg.ChecksumArrayLength != 0:
			buff.seek(msg.ChecksumArrayOffset)
			msg.ChecksumValue = buff.read(msg.ChecksumArrayLength)

		return msg

class MESSAGE_HEADER:
	#define MESSAGE_SIGNATURE    0x535458454f47454ei64
	def __init__(self):
		self.Signature = b'NEGOEXTS'
		self.MessageType = None #MESSAGE_TYPE enum
		self.SequenceNum = None #ULONG  the message sequence number of this, conversation, starting with 0 and sequentially incremented
		self.cbHeaderLength = None #ULONG // the header length of this message, including the message specific header, excluding the payload
		self.cbMessageLength = None #ULONG // the length of this message
		self.ConversationId = None #CONVERSATION_ID
		self._hdrsize = 40

	def to_bytes(self):
		t = self.Signature
		t += self.MessageType.value.to_bytes(4, byteorder='little', signed = False)
		t += self.SequenceNum.to_bytes(4, byteorder='little', signed = False)
		t += self.cbHeaderLength.to_bytes(4, byteorder='little', signed = False)
		t += self.cbMessageLength.to_bytes(4, byteorder='little', signed = False)
		t += self.ConversationId
		return t
	
	@staticmethod
	def from_bytes(data):
		return MESSAGE_HEADER.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		msg = MESSAGE_HEADER()
		msg.Signature = buff.read(8)
		msg.MessageType = MESSAGE_TYPE(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		msg.SequenceNum = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.cbHeaderLength = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.cbMessageLength = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.ConversationId = buff.read(16) # guid, but basically doesnt matter
		return msg

class NEGO_MESSAGE:
	def __init__(self):
		self.Header = None #MESSAGE_HEADER
		self.Random = os.urandom(32) # 32 bytes
		self.ProtocolVersion = 0 #ULONG64  // version of the protocol, this contains 0
		self.AuthSchemeArrayOffset = 0 #ULONG // each element contains an AUTH_SCHEME
		self.AuthSchemeCount = 0 #ULONG
		self.ExtensionArrayOffset = 0 #ULONG // each element contains an AUTH_SCHEME
		self.ExtensionCount = 0 #ULONG
		self.msgsize = 40+32+8+4+4+4+4
		
		self.Extensions = []
		self.AuthSchemes = [NEGOEXTS_DEFAULT_AUTHID]
	
	def to_bytes(self):
		self.Header.cbHeaderLength = self.msgsize
		self.AuthSchemeArrayOffset = self.msgsize
		buff = io.BytesIO()
		for auth in self.AuthSchemes:
			buff.write(auth.bytes_le)
			self.AuthSchemeCount += 1
		buff.seek(0)
		payload = buff.read()
		buff = io.BytesIO()
		if len( self.Extensions) > 0:
			self.ExtensionArrayOffset = self.msgsize + len(payload)
			for ext in self.Extensions:
				ext.to_buffer(buff)
				self.ExtensionCount += 1
			
		buff.seek(0)
		payload += buff.read()

		self.Header.cbMessageLength = self.msgsize + len(payload)
		

		res = self.Header.to_bytes()
		res += self.Random
		res += self.ProtocolVersion.to_bytes(8, byteorder='little', signed = False)
		res += self.AuthSchemeArrayOffset.to_bytes(4, byteorder='little', signed = False)
		res += self.AuthSchemeCount.to_bytes(4, byteorder='little', signed = False)
		res += self.ExtensionArrayOffset.to_bytes(4, byteorder='little', signed = False)
		res += self.ExtensionCount.to_bytes(4, byteorder='little', signed = False)
		res += payload

		return res
	
	@staticmethod
	def from_bytes(data):
		return NEGO_MESSAGE.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		start_offset = buff.tell()
		msg = NEGO_MESSAGE()
		msg.Header = MESSAGE_HEADER.from_buffer(buff)
		msg.Random = buff.read(32)
		msg.ProtocolVersion = int.from_bytes(buff.read(8), byteorder='little', signed = False)
		msg.AuthSchemeArrayOffset = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.AuthSchemeCount = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.ExtensionArrayOffset = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.ExtensionCount = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		
		msg.AuthSchemes = []
		if msg.AuthSchemeCount != 0:
			buff.seek(start_offset + msg.AuthSchemeArrayOffset)
			for i in range(msg.AuthSchemeCount):
				auth_data = buff.read(16) # AUTH_SCHEME is a GUID
				msg.AuthSchemes.append(uuid.UUID(bytes_le=auth_data))

		if msg.ExtensionCount != 0:
			buff.seek(start_offset + msg.ExtensionArrayOffset)
			ext_data = EXTENSION.from_buffer(buff)
			msg.Extensions.append(ext_data)


		#msg.AuthSchemes = AUTH_SCHEME_VECTOR.from_buffer(buff)
		#msg.Extensions = EXTENSION_VECTOR.from_buffer(buff)
		return msg


class EXCHANGE_MESSAGE:
	def __init__(self):
		self.Header = None #MESSAGE_HEADER // MESSAGE_TYPE_CHALLENGE for the acceptor, or MESSAGE_TYPE_AP_REQUEST for the initiator MESSAGE_TYPE_INITiATOR_META_DATA for the initiator metadata MESSAGE_TYPE_ACCEPTOR_META_DATA for the acceptor metadata
		self.AuthScheme = None #AUTH_SCHEME
		self.ExchangeOffset = None
		self.ExchangeByteCount = None
		self.Exchange = None #BYTE_VECTOR  // contains the opaque handshake message for the authentication scheme
		self.msgsize = 40 + 16 + 4 + 4
		self.exchange_data_raw = None

	def to_bytes(self):
		self.ExchangeByteCount = len(self.Exchange)
		self.ExchangeOffset = self.msgsize
		self.Header.cbMessageLength = self.msgsize + self.ExchangeByteCount
		self.Header.cbHeaderLength = self.msgsize

		res = self.Header.to_bytes()
		res += self.AuthScheme.bytes_le
		res += self.ExchangeOffset.to_bytes(4, byteorder='little', signed = False)
		res += self.ExchangeByteCount.to_bytes(4, byteorder='little', signed = False)
		res += self.Exchange

		return res


	@staticmethod
	def from_bytes(data):
		return EXCHANGE_MESSAGE.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		start_offset = buff.tell()
		msg = EXCHANGE_MESSAGE()
		msg.Header = MESSAGE_HEADER.from_buffer(buff)
		msg.AuthScheme = uuid.UUID(bytes_le=buff.read(16))
		msg.ExchangeOffset = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.ExchangeByteCount = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		buff.seek(start_offset + msg.ExchangeOffset)
		exch_data = buff.read(msg.ExchangeByteCount)
		msg.exchange_data_raw = exch_data
		
		if msg.Header.MessageType in [MESSAGE_TYPE.AP_REQUEST, MESSAGE_TYPE.CHALLENGE]:
			token = PKU2U_TOKEN.from_bytes(exch_data)
			msg.Exchange = token
		
		elif msg.Header.MessageType in [MESSAGE_TYPE.INITIATOR_META_DATA, MESSAGE_TYPE.ACCEPTOR_META_DATA]:
			token = MetaData.load(exch_data)
			#x = X.load(token.native['1'][0]['data'])
			msg.Exchange = token
		
		else:
			msg.Exchange = exch_data
		return msg


class VERIFY_MESSAGE:
	def __init__(self):
		self.Header = None # MESSAGE_HEADER  // MESSAGE_TYPE_VERIFY
		self.AuthScheme = None #AUTH_SCHEME

		self.cbHeaderLength = 20 #ULONG, always 20
		self.ChecksumScheme = 1 #ULONG RFC3961
		self.ChecksumType = None #ULONG // in the case of RFC3961 scheme, this is the RFC3961 checksum type
		self.ChecksumArrayOffset = None
		self.ChecksumArrayLength = None
		self.Checksum = None
		self.msgsize = 40 + 16 + 20 + 4
	
	def to_bytes(self):
		self.ChecksumArrayLength = len(self.Checksum)
		self.ChecksumArrayOffset = self.msgsize
		self.Header.cbMessageLength = self.msgsize + len(self.Checksum)
		self.Header.cbHeaderLength = self.msgsize

		res = self.Header.to_bytes()
		res += self.AuthScheme.bytes_le
		res += self.cbHeaderLength.to_bytes(4, byteorder='little', signed = False)
		res += self.ChecksumScheme.to_bytes(4, byteorder='little', signed = False)
		res += self.ChecksumType.to_bytes(4, byteorder='little', signed = False)
		res += self.ChecksumArrayOffset.to_bytes(4, byteorder='little', signed = False)
		res += self.ChecksumArrayLength.to_bytes(4, byteorder='little', signed = False)
		res += b'\x00' * 4 # manually adding pad
		res += self.Checksum

		return res
	
	@staticmethod
	def from_bytes(data):
		return VERIFY_MESSAGE.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		start_offset = buff.tell()
		msg = VERIFY_MESSAGE()
		msg.Header = MESSAGE_HEADER.from_buffer(buff)
		msg.AuthScheme = uuid.UUID(bytes_le=buff.read(16))

		msg.cbHeaderLength = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.ChecksumScheme = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.ChecksumType = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.ChecksumArrayOffset = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.ChecksumArrayLength = int.from_bytes(buff.read(4), byteorder='little', signed = False)

		if msg.ChecksumArrayLength != 0:
			buff.seek(msg.ChecksumArrayOffset)
			msg.Checksum = buff.read(msg.ChecksumArrayLength)
		return msg


class ALERT_MESSAGE:
	def __init__(self):
		self.Header = None #MESSAGE_HEADER
		self.AuthScheme = None #AUTH_SCHEME
		self.ErrorCode = None# ULONG // an NTSTATUS code
		self.AlertArrayOffset = None
		self.AlertCount = None

		self.Alerts = []

	@staticmethod
	def from_bytes(data):
		return ALERT_MESSAGE.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		start_offset = buff.tell()
		msg = ALERT_MESSAGE()
		msg.Header = MESSAGE_HEADER.from_buffer(buff)
		msg.AuthScheme = uuid.UUID(bytes_le=buff.read(16))
		msg.ErrorCode = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.AlertArrayOffset = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.AlertCount = int.from_bytes(buff.read(4), byteorder='little', signed = False)

		if msg.AlertCount != 0:
			buff.seek(start_offset + msg.AlertArrayOffset)
			for _ in range(msg.AlertCount):
				msg.Alerts.append(Alert.from_buffer(buff))


		return msg

class ALERT_TYPE(enum.Enum):
	PULSE = 1

class ALERT_VERIFY(enum.Enum):
	NO_KEY = 1

class ALERT_PULSE:
	def __init__(self):
		self.cbHeaderLength = None
		self.Reason = None

class ALERT:
	def __init__(self):
		self.AlertType = None # MESSAGE_HEADER  // MESSAGE_TYPE_VERIFY
		self.AlertDataOffset = None
		self.AlertDataLength = None

		self.AlertValue = None # BYTE_VECTOR //alert types
	
	@staticmethod
	def from_bytes(data):
		return ALERT.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		start_offset = buff.tell()
		msg = ALERT()
		msg.AlertType = ALERT_TYPE(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		msg.AlertDataOffset = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.AlertDataLength = int.from_bytes(buff.read(4), byteorder='little', signed = False)

		if msg.AlertDataLength != 0:
			buff.seek(msg.AlertDataOffset)
			msg.AlertValue = buff.read(msg.AlertDataLength)
		
		return msg


def generate_initiator_metadata(seq_number, conv_id, metadata, authscheme = NEGOEXTS_DEFAULT_AUTHID):
	hdr = MESSAGE_HEADER()
	hdr.MessageType = MESSAGE_TYPE.INITIATOR_META_DATA #MESSAGE_TYPE enum
	hdr.SequenceNum = seq_number
	hdr.ConversationId = conv_id

	em = EXCHANGE_MESSAGE()
	em.Header = hdr
	em.AuthScheme = authscheme
	em.Exchange = metadata

	return em.to_bytes()

def generate_init_nego(seq_number, conv_id, authschemes = None, extensions = None):
	hdr = MESSAGE_HEADER()
	hdr.MessageType = MESSAGE_TYPE.INITIATOR_NEGO
	hdr.SequenceNum = seq_number
	hdr.ConversationId = conv_id

	nego = NEGO_MESSAGE()
	nego.Header = hdr
	if authschemes is not None:
		nego.AuthSchemes = authschemes
	if extensions is not None:
		nego.Extensions = extensions

	return nego.to_bytes()

def generate_ap_req(seq_number, conv_id, ap_req, tok_type, authscheme = NEGOEXTS_DEFAULT_AUTHID):
	hdr = MESSAGE_HEADER()
	hdr.MessageType = MESSAGE_TYPE.AP_REQUEST
	hdr.SequenceNum = seq_number
	hdr.ConversationId = conv_id

	exchange = EXCHANGE_MESSAGE()
	exchange.Header = hdr
	exchange.AuthScheme = authscheme
	token = PKU2U_TOKEN()
	token.tok_id = tok_type #PKU2U_TOKEN_TYPE.KRB_AS_REQ
	token.inner_token = ap_req
	exchange.Exchange = token.to_bytes()
	return exchange.to_bytes(), token.to_bytes()

def generate_verify(seq_number, conv_id, checksum, checksumtype, authscheme = NEGOEXTS_DEFAULT_AUTHID):
	hdr = MESSAGE_HEADER()
	hdr.MessageType = MESSAGE_TYPE.VERIFY
	hdr.SequenceNum = seq_number
	hdr.ConversationId = conv_id

	verify = VERIFY_MESSAGE()
	verify.Header = hdr
	verify.AuthScheme = authscheme
	verify.Checksum = checksum
	verify.ChecksumType = checksumtype


	return verify.to_bytes()

def negoexts_parse_bytes(data):
	return negoexts_parse_buffer(io.BytesIO(data))

def negoexts_parse_buffer(buff):
	resd = {}
	start = buff.tell()
	buff.seek(0,2)
	end = buff.tell()
	buff.seek(start)
	maxiter = 255
	while buff.tell() != end and maxiter != 0:		
		maxiter -= 1
		start = buff.tell()
		hdr = MESSAGE_HEADER.from_buffer(buff)
		buff.seek(start)
		data = buff.read(hdr.cbMessageLength)
		if hdr.MessageType in [MESSAGE_TYPE.INITIATOR_NEGO, MESSAGE_TYPE.ACCEPTOR_NEGO]:
			res = NEGO_MESSAGE.from_bytes(data)
		elif hdr.MessageType in [MESSAGE_TYPE.CHALLENGE, MESSAGE_TYPE.INITIATOR_META_DATA, MESSAGE_TYPE.ACCEPTOR_META_DATA, MESSAGE_TYPE.AP_REQUEST]:
			res = EXCHANGE_MESSAGE.from_bytes(data)
		elif hdr.MessageType == MESSAGE_TYPE.VERIFY:
			res = VERIFY_MESSAGE.from_bytes(data)
		elif hdr.MessageType == MESSAGE_TYPE.ALERT:
			res = ALERT_MESSAGE.from_bytes(data)
		
		resd[hdr.MessageType] = res

	if maxiter == 0:
		print('SAD')

	return resd